"""
    Module for the Partition configuration nodes and its dependencies.
"""

import abc
import collections
import functools
import json
import typing

import nrrd
import numpy as np

from .. import config
from ..config import types
from ..exceptions import (
    ConfigurationError,
    LayoutError,
    NodeNotFoundError,
    RequirementError,
)
from ..storage._chunks import Chunk
from ..storage._files import NrrdDependencyNode
from ..storage._util import _cached_file
from ..voxels import VoxelSet
from ._layout import Layout, RhomboidData

if typing.TYPE_CHECKING:
    from ..core import Scaffold


def _size_requirements(section):
    if "thickness" not in section and "volume_scale" not in section:
        raise RequirementError(
            "Either a `thickness` or `volume_scale` attribute required"
        )


class _backref_property(property):
    def __backref__(self, instance, value):
        setattr(instance, "_region", value)


@config.dynamic(
    attr_name="type",
    required=False,
    default="layer",
    auto_classmap=True,
)
class Partition(abc.ABC):
    scaffold: "Scaffold"
    name: str = config.attr(key=True)

    @_backref_property
    def region(self):
        return self._region

    @property
    def data(self):
        # The data property is read-only to users, but `_data` is assigned
        # during the layout process
        return self._data

    @abc.abstractmethod
    def volume(self, chunk=None):  # pragma: nocover
        """
        Calculate the volume of the partition in μm^3.

        :param chunk: If given, limit the volume of the partition inside of the chunk.
        :type chunk: bsb.storage._chunks.Chunk
        :returns: Volume of the partition (in the chunk)
        :rtype: float
        """
        pass

    @abc.abstractmethod
    def surface(self, chunk=None):  # pragma: nocover
        """
        Calculate the surface of the partition in μm^2.

        :param chunk: If given, limit the surface of the partition inside of the chunk.
        :type chunk: bsb.storage._chunks.Chunk
        :returns: Surface of the partition (in the chunk)
        :rtype: float
        """
        pass

    @abc.abstractmethod
    def to_chunks(self, chunk_size):  # pragma: nocover
        """
        Calculate all the chunks this partition occupies when cut into ``chunk_sized``
        pieces.

        :param chunk_size: Size per chunk (in μm). The slicing always starts at [0, 0, 0].
        :type chunk_size: numpy.ndarray
        :returns: Chunks occupied by this partition
        :rtype: List[bsb.storage._chunks.Chunk]
        """
        pass

    @abc.abstractmethod
    def chunk_to_voxels(self, chunk):  # pragma: nocover
        """
        Voxelize the partition's occupation in this chunk. Required to fill the partition
        with cells by the placement module.

        :param chunk: The chunk to calculate voxels for.
        :type chunk: bsb.storage._chunks.Chunk
        :returns: The set of voxels that together make up the shape of this partition in
          this chunk.
        :rtype: bsb.voxels.VoxelSet
        """
        pass

    @abc.abstractmethod
    def rotate(self, rotation):  # pragma: nocover
        """
        Rotate the partition by the given rotation object.

        :param rotation: Rotation object.
        :type rotation: scipy.spatial.transform.Rotation
        :raises: :class:`.exceptions.LayoutError` if the rotation needs to be rejected.
        """
        pass

    @abc.abstractmethod
    def translate(self, offset):  # pragma: nocover
        """
        Translate the partition by the given offset.

        :param offset: Offset, XYZ.
        :type offset: numpy.ndarray
        :raises: :class:`.exceptions.LayoutError` if the translation needs to be rejected.
        """
        pass

    @abc.abstractmethod
    def scale(self, factors):  # pragma: nocover
        """
        Scale up/down the partition according to the given factors.

        :param factors: Scaling factors, XYZ.
        :type factors: numpy.ndarray
        :raises: :class:`.exceptions.LayoutError` if the scaling needs to be rejected.
        """
        pass

    @abc.abstractmethod
    def get_layout(self, hint):  # pragma: nocover
        """
        Given a Layout as hint to begin from, create a Layout object that describes how
        this partition would like to be laid out.

        :param hint: The layout space that this partition should place itself in.
        :type hint: bsb.topology._layout.Layout
        :returns: The layout describing the space this partition takes up.
        :rtype: bsb.topology._layout.Layout
        """
        pass


@config.node
class Rhomboid(Partition, classmap_entry="rhomboid"):
    dimensions: list[float] = config.attr(type=types.list(type=float, size=3))
    can_scale: bool = config.attr(type=bool, default=True)
    origin: list[float] = config.attr(type=types.list(type=float, size=3))
    can_move: bool = config.attr(type=bool, default=True)
    orientation: list[float] = config.attr(type=types.list(type=float, size=3))
    can_rotate: bool = config.attr(type=bool, default=True)

    def volume(self, chunk=None):
        if chunk is not None:
            # Create an intersection between the partition and the chunk
            low = np.maximum(self.ldc, chunk.ldc)
            high = np.minimum(self.mdc, chunk.mdc)
            return np.prod(np.maximum(high - low, 0))
        else:
            return np.prod(self.data.dimensions)

    @property
    def mdc(self):
        return self._data.mdc

    @property
    def ldc(self):
        return self._data.ldc

    def surface(self, chunk=None):
        if chunk is not None:
            # Gets the xz "square" from a volume
            sq = lambda v: np.array(v)[[0, 1]]
            ldc = sq(self.ldc)
            mdc = sq(self.mdc)
            cl = sq(chunk.ldc)
            cm = sq(chunk.mdc)
            # Create an intersection between the partition and the chunk
            low = np.maximum(ldc, cl)
            high = np.minimum(mdc, cm)
            return np.prod(np.maximum(high - low, 0))
        else:
            return self.data.width * self.data.depth

    def to_chunks(self, chunk_size):
        # Get the low and high range of the data in chunk coordinates
        low_r = np.floor(self.ldc / chunk_size).astype(int)
        high_r = np.ceil(self.mdc / chunk_size).astype(int)
        # Create a grid that includes all the chunk coordinates within those data
        coords = np.mgrid[tuple(range(low, high) for low, high in zip(low_r, high_r))]
        # Order the coordinate grid into a list of chunk coordinates.
        return np.column_stack(tuple(dim.ravel() for dim in coords))

    def chunk_to_voxels(self, chunk):
        """
        Return an approximation of this partition intersected with a chunk as a list of
        voxels.

        Default implementation creates a parallellepepid intersection between the
        LDC, MDC and chunk data.
        """
        low = np.maximum(self.ldc, chunk.ldc)
        high = np.minimum(self.mdc, chunk.mdc)
        # Return 0 voxels when the coords are OOB for this partition
        if np.any(low > high):
            return VoxelSet.empty()
        else:
            return VoxelSet.one(low, high)

    def rotate(self, rot):
        raise LayoutError("Rotation not implemented yet.")

    def translate(self, translation):
        self.data.ldc += translation
        self.data.mdc += translation

    def scale(self, factors):
        self.data.mdc = self.data.ldc + (self.data.mdc - self.data.ldc) * factors

    def get_dependencies(self):
        """
        Return other partitions or regions that need to be laid out before this.
        """
        return []

    def get_layout(self, hint):
        if self.dimensions is None:
            dim = hint.data.mdc - hint.data.ldc
        else:
            dim = self.dimensions
        if self.origin is None:
            orig = hint.data.ldc.copy()
        else:
            orig = self.origin
        return Layout(RhomboidData(orig, dim + orig), owner=self)


@config.node
class Layer(Rhomboid, classmap_entry="layer"):
    dimensions = config.unset()
    thickness: float = config.attr(type=float, required=_size_requirements)
    volume_scale: list[float] = config.attr(
        type=types.or_(
            types.list(float, size=2),
            types.scalar_expand(
                float,
                lambda x: [x, x],
            ),
        ),
        default=lambda: [1.0, 1.0],
        call_default=True,
    )
    axis: typing.Union[typing.Literal["x"], typing.Literal["y"], typing.Literal["z"]] = (
        config.attr(type=types.in_(["x", "y", "z"]), default="z")
    )
    stack_index: float = config.attr(type=float, default=0)

    def get_layout(self, hint):
        axis = ["x", "y", "z"].index(self.axis)
        dim = hint.data.mdc - hint.data.ldc
        dim[axis] = self.thickness
        if self.origin is None:
            orig = hint.data.ldc.copy()
        else:
            orig = self.origin
        return Layout(RhomboidData(orig, dim + orig), owner=self)

    # TODO: Layer scaling


@config.node
class Voxels(Partition, abc.ABC, classmap_entry=None):
    """
    Partition based on a set of voxels.
    """

    @abc.abstractmethod
    def get_voxelset(self):
        pass

    @functools.cached_property
    def voxelset(self):
        return self.get_voxelset()

    def to_chunks(self, chunk_size):
        return self.voxelset.snap_to_grid(chunk_size, unique=True)

    def _lookup(self, chunk):
        if not hasattr(self, "_map"):
            vs = self.voxelset.snap_to_grid(chunk.dimensions)
            map = {}
            for i, idx in enumerate(vs):
                map.setdefault(idx.view(Chunk), []).append(i)
            self._map = {k: self.voxelset[v] for k, v in map.items()}
        return self._map

    def chunk_to_voxels(self, chunk):
        return self._lookup(chunk).get(chunk, VoxelSet.empty())

    def get_layout(self, hint):
        return Layout(RhomboidData(*self.voxelset.bounds), owner=self)

    def rotate(self, rotation):
        raise LayoutError("Axis-aligned voxelsets can't be rotated.")

    def scale(self, factor):
        raise LayoutError("Voxelset scaling not supported.")

    def translate(self, offset):
        raise LayoutError("Voxelset translation not supported.")

    def surface(self, chunk=None):
        raise LayoutError("Voxelset surface calculations not supported.")

    def volume(self, chunk=None):
        if chunk is not None:
            vs = self.chunk_to_voxels(chunk)
        else:
            vs = self.voxelset
        return vs.volume


@config.node
class NrrdVoxels(Voxels, classmap_entry="nrrd"):
    """
    Voxel partition whose voxelset is loaded from an NRRD file. By default it includes all the
    nonzero voxels in the file, but other masking conditions can be specified. Additionally, data
    can be associated to each voxel by inclusion of (multiple) source NRRD files.
    """

    source: NrrdDependencyNode = config.attr(
        type=NrrdDependencyNode,
        required=types.mut_excl("source", "sources", required=False),
    )
    """Path to the NRRD file containing volumetric data to associate with the partition.
    If source is set, then sources should not be set."""
    sources: NrrdDependencyNode = config.list(
        type=NrrdDependencyNode,
        required=types.mut_excl("source", "sources", required=False),
    )
    """List of paths to NRRD files containing volumetric data to associate with the Partition.
    If sources is set, then source should not be set."""
    mask_value: int = config.attr(type=int)
    """Integer value to filter in mask_source (if it is set, otherwise sources/source) to create a 
    mask of the voxel set(s) used as input."""
    mask_source: NrrdDependencyNode = config.attr(type=NrrdDependencyNode)
    """Path to the NRRD file containing the volumetric annotation data of the Partition."""
    mask_only: bool = config.attr(type=bool, default=False)
    """Flag to indicate no voxel data needs to be stored"""
    voxel_size: int = config.attr(type=int, required=True)
    """Size of each voxel."""
    keys: list[str] = config.attr(type=types.list(str))
    """List of names to assign to each source of the Partition."""
    sparse: bool = config.attr(type=bool, default=True)
    """
    Boolean flag to expect a sparse or dense mask. If the mask selects most
    voxels, use ``dense``, otherwise use ``sparse``.
    """
    strict: bool = config.attr(type=bool, default=True)
    """Boolean flag to check the sources and the mask sizes. 
    When the flag is True, sources and mask should have exactly the same sizes;
    otherwise, sources sizes should be greater than mask sizes."""

    def get_mask(self):
        """
        Get the mask to apply on the sources' data of the partition.

        :returns: A tuple of arrays, one for each dimension of the mask, containing the indices of
            the non-zero elements in that dimension.
        """

        mask_shape = self._validate()
        mask = np.zeros(mask_shape, dtype=bool)
        if self.sparse:
            # Use integer (sparse) indexing
            mask = [np.empty((0,), dtype=int) for i in range(3)]
            for mask_src in self._mask_src:
                mask_data = mask_src.get_data()
                new_mask = np.nonzero(self._mask_cond(mask_data))
                for i, mask_vector in enumerate(new_mask):
                    mask[i] = np.concatenate((mask[i], mask_vector))
            inter = np.unique(mask, axis=1)
            mask = tuple(inter[i, :] for i in range(3))
        else:
            # Use boolean (dense) indexing
            for mask_src in self._mask_src:
                mask_data = mask_src.get_data()
                mask = mask | self._mask_cond(mask_data)
            mask = np.nonzero(mask)
        return mask

    def get_voxelset(self):
        """
        Creates a VoxelSet of the sources of the Partition that matches its mask.

        :returns: VoxelSet of the Partition sources.
        """

        mask = self.get_mask()
        voxel_data = None
        if not self.mask_only:
            voxel_data = np.empty((len(mask[0]), len(self._src)))
            for i, source in enumerate(self._src):
                voxel_data[:, i] = source.get_data()[mask]

        return VoxelSet(
            np.transpose(mask),
            self.voxel_size,
            data=voxel_data,
            data_keys=self.keys,
        )

    def _validate(self):
        self._validate_sources()
        shape = self._validate_source_compat()
        self._validate_mask_condition()
        return shape

    def _validate_sources(self):
        if self.source is not None:
            self._src = [self.source]
        else:
            self._src = self.sources.copy()
        if self.mask_source is not None:
            self._mask_src = [self.mask_source]
        else:
            self._mask_src = self._src.copy()

    def _validate_source_compat(self):
        mask_headers = {s: s.get_header() for s in self._mask_src}
        source_headers = {s: s.get_header() for s in self._src}
        all_headers = mask_headers.copy()
        all_headers.update(source_headers)
        dim_probs = [(s, d) for s, h in all_headers.items() if (d := h["dimension"]) != 3]
        if dim_probs:
            summ = ", ".join(f"'{s}' has {d}" for s, d in dim_probs)
            raise ConfigurationError(f"NRRD voxels must contain 3D arrays; {summ}")
        mask_sizes = {s: [*h["sizes"]] for s, h in mask_headers.items()}
        source_sizes = {s: [*h["sizes"]] for s, h in source_headers.items()}
        all_sizes = mask_sizes.copy()
        all_sizes.update(source_sizes)
        mask_shape = np.maximum.reduce([*mask_sizes.values()])
        if self.mask_only:
            src_shape = mask_shape
        else:
            src_shape = np.minimum.reduce([*source_sizes.values()])
        first = _repeat_first()
        # Check for any size mismatch
        if self.strict and any(size != first(size) for size in all_sizes.values()):
            raise ConfigurationError(
                f"NRRD file size mismatch in `{self.get_node_name()}`: {all_sizes}"
            )
        elif np.any(mask_shape > src_shape):
            raise ConfigurationError(
                f"NRRD mask too big; it may select OOB source voxels:"
                + f" {mask_shape} > {src_shape}"
            )
        return mask_shape

    def _validate_mask_condition(self):
        if self.mask_value:
            self._mask_cond = lambda data: data == self.mask_value
        else:
            self._mask_cond = lambda data: data != 0


@config.node
class AllenStructure(NrrdVoxels, classmap_entry="allen"):
    """
    Partition based on the Allen Institute for Brain Science mouse brain region ontology, later
    referred as Allen Mouse Brain Region Hierarchy (AMBRH)
    """

    struct_id: int = config.attr(
        type=int, required=types.mut_excl("struct_id", "struct_name", required=False)
    )
    """Id of the region to filter within the annotation volume according to the AMBRH.
    If struct_id is set, then struct_name should not be set."""
    struct_name: str = config.attr(
        type=types.str(strip=True, lower=True),
        required=types.mut_excl("struct_id", "struct_name", required=False),
        key=True,
    )
    """Name or acronym of the region to filter within the annotation volume according to the AMBRH.
    If struct_name is set, then struct_id should not be set."""

    @config.property(type=int)
    def voxel_size(self):
        """Size of each voxel."""
        return self._voxel_size if self._voxel_size is not None else 25

    @voxel_size.setter
    def voxel_size(self, value):
        self._voxel_size = value

    @config.property(type=bool)
    def mask_only(self):
        return self.source is None and len(self.sources) == 0

    @config.property(type=str)
    @functools.cache
    def mask_source(self):
        if hasattr(self, "_annotations_file"):
            return self._annotations_file
        else:
            node = NrrdDependencyNode()
            node._file = _cached_file(
                "http://download.alleninstitute.org/informatics-archive/current-release/mouse_ccf/annotation/ccf_2017/annotation_25.nrrd",
            )
            return node

    @mask_source.setter
    def mask_source(self, value):
        self._mask_source = value
        if value is not None:
            self._annotations_file = NrrdDependencyNode(file=value)
        elif hasattr(self, "_annotations_file"):
            delattr(self, "_annotations_file")

    @classmethod
    @functools.cache
    def _dl_structure_ontology(cls):
        return json.loads(
            _cached_file(
                "http://api.brain-map.org/api/v2/structure_graph_download/1.json"
            ).get_content()[0]
        )["msg"]

    @classmethod
    def get_structure_mask_condition(cls, find):
        """
        Return a lambda that when applied to the mask data, returns a mask that delineates
        the Allen structure.

        :param find: Acronym, Name or ID of the Allen structure.
        :type find: Union[str, int]
        :returns: Masking lambda
        :rtype: Callable[numpy.ndarray]
        """
        mask = cls.get_structure_idset(find)
        if len(mask) > 1:
            return lambda data: np.isin(data, mask)
        else:
            mask0 = mask[0]
            return lambda data: data == mask0

    @classmethod
    def get_structure_mask(cls, find):
        """
        Returns the mask data delineated by the Allen structure.

        :param find: Acronym, Name or ID of the Allen structure.
        :type find: Union[str, int]
        :returns: A boolean of the mask filtered based on the Allen structure.
        :rtype: Callable[numpy.ndarray]
        """
        mask_data, _ = nrrd.read(cls._dl_mask())
        return cls.get_structure_mask_condition(find)(mask_data)

    @classmethod
    def get_structure_idset(cls, find):
        """
        Return the set of IDs that make up the requested Allen structure.

        :param find: Acronym or ID of the Allen structure.
        :type find: Union[str, int]
        :returns: Set of IDs
        :rtype: numpy.ndarray
        """
        struct = cls.find_structure(find)
        values = set()

        def flatmask(item):
            values.add(item["id"])

        cls._visit_structure([struct], flatmask)
        return np.array([*values], dtype=int)

    @classmethod
    def find_structure(cls, id):
        """
        Find an Allen structure by name, acronym or ID.

        :param id: Query for the name, acronym or ID of the Allen structure.
        :type id: Union[str, int, float]
        :returns: Allen structure node of the Allen ontology tree.
        :rtype: dict
        :raises: NodeNotFoundError
        """
        if isinstance(id, str):
            treat = lambda s: s.strip().lower()
            name = treat(id)
            find = lambda x: treat(x["name"]) == name or treat(x["acronym"]) == name
        elif isinstance(id, int) or isinstance(id, float):
            id = int(id)
            find = lambda x: x["id"] == id
        else:
            raise TypeError(f"Argument must be a string or a number. {type(id)} given.")
        try:
            return cls._find_structure(find)
        except NodeNotFoundError:
            raise NodeNotFoundError(f"Could not find structure '{id}'") from None

    @classmethod
    def _find_structure(cls, find):
        result = None

        def visitor(item):
            nonlocal result
            if find(item):
                result = item
                return True

        tree = cls._dl_structure_ontology()
        cls._visit_structure(tree, visitor)
        if result is None:
            raise NodeNotFoundError("Could not find a node that satisfies constraints.")
        return result

    @classmethod
    def _visit_structure(cls, tree, visitor):
        deck = collections.deque(tree)
        while True:
            try:
                item = deck.popleft()
            except IndexError:
                break
            if visitor(item):
                break
            deck.extend(item["children"])

    def _validate_mask_condition(self):
        # We override the `NrrdVoxels`' `_validate_mask_condition` and use this
        # function as a hook to find and set the mask condition to select every voxel that
        # has an id that is part of the structure.
        id = self.struct_id if self.struct_id is not None else self.struct_name
        self._mask_cond = self.get_structure_mask_condition(id)


def _safe_hread(s):
    try:
        return nrrd.read_header(s)
    except StopIteration:
        raise IOError(f"Empty NRRD file '{s}' could not be read.") from None


def _repeat_first():
    _set = False
    first = None

    def repeater(val):
        nonlocal _set, first
        if not _set:
            first, _set = val, True
        return first

    return repeater


__all__ = ["AllenStructure", "Layer", "NrrdVoxels", "Partition", "Rhomboid", "Voxels"]



def _dunder(*hooks):
    return "__" + "_".join(hooks) + "__"


def run_hook(obj, hook, *args, **kwargs):
    """
    Execute the ``hook`` hook of ``obj``.

    Runs the ``hook`` method ``obj`` but also looks through the class hierarchy for
    essential hooks with the name ``__<hook>__``.

    .. note::

      Essential hooks are only ran if the method is called using ``run_hook`` while
      non-essential hooks are wrapped around the method and will always be executed when
      the method is called (see https://github.com/dbbs-lab/bsb/issues/158).
    """
    # Traverse the MRO tree and execute all essential (__hook__) methods.
    for parent in reversed(obj.__class__.__mro__):
        if hasattr(parent, _dunder(hook)):
            getattr(parent, _dunder(hook))(obj, *args, **kwargs)
    # Execute the final regular hook method.
    if hasattr(obj, hook):
        getattr(obj, hook)(*args, **kwargs)


def on(hook, cls, essential=False, before=False):
    """
    Register a class hook.

    :param hook: Name of the method to hook.
    :type hook: str
    :param cls: Class to hook.
    :type cls: type
    :param essential: If the hook is essential, it will always be executed even in child
      classes that override the hook. Essential hooks are only lost if the method on
      ``cls`` is replaced.
    :type essential: bool
    :param before: If ``before`` the hook is executed before the method, otherwise
      afterwards.
    :type before: bool
    """
    if essential:
        hook = _dunder(hook)
    return _hook(cls, hook, before)


def after(hook, cls, essential=False):
    """
    Register a class hook to run after the target method.

    :param hook: Name of the method to hook.
    :type hook: str
    :param cls: Class to hook.
    :type cls: type
    :param essential: If the hook is essential, it will always be executed even in child
      classes that override the hook. Essential hooks are only lost if the method on
      ``cls`` is replaced.
    :type essential: bool
    """
    return on(hook, cls, essential=essential, before=False)


def before(hook, cls, essential=False):
    """
    Register a class hook to run before the target method.

    :param hook: Name of the method to hook.
    :type hook: str
    :param cls: Class to hook.
    :type cls: type
    :param essential: If the hook is essential, it will always be executed even in child
      classes that override the hook. Essential hooks are only lost if the method on
      ``cls`` is replaced.
    :type essential: bool
    """
    return on(hook, cls, essential=essential, before=True)


def _super(cls, attr):
    for c in list(cls.__mro__)[1:]:
        try:
            return getattr(c, attr)
        except:
            pass


def has_hook(instance, hook):
    """
    Checks the existence of a method or essential method on the ``instance``.

    :param instance: Object to inspect.
    :type instance: object
    :param hook: Name of the hook to look for.
    :type hook: str
    """
    return hasattr(instance, hook) or hasattr(instance, _dunder(hook))


def overrides(cls, hook, mro=False):
    """
    Returns ``True`` if a class has implemented a method or ``False`` if it has inherited
    it.

    :param cls: Class to inspect.
    :type cls: class
    :param hook: Name of the hook to look for.
    :type hook: str
    """
    if not mro:
        return hook in vars(cls)
    else:

        class NotDefined:
            pass

        return getattr(cls, hook, NotDefined) is not getattr(object, hook, NotDefined)


def _hook(cls, hook, before):
    # Returns a decorator that will wrap the existing implementation if it exists,
    # otherwise returns a decorator that will simply add the decorated function to the
    # class.
    if overrides(cls, hook):
        return _make_wrapper(cls, hook, before)
    else:
        return _make_injector(cls, hook)


def _make_injector(cls, hook):
    def decorator(func):
        # Set the decorated function as the hooked function.
        _set_hooked_func(cls, hook, func)

    return decorator


def _get_func_to_hook(cls, hook):
    return getattr(cls, hook)


def _set_hooked_func(cls, hook, hooked):
    return setattr(cls, hook, hooked)


def _make_wrapper(cls, hook, before):
    func_to_hook = _get_func_to_hook(cls, hook)

    def decorator(func):
        if before:
            # Wrapper that executes the decorated function before the class method
            def hooked(*args, **kwargs):
                func(*args, **kwargs)
                r = func_to_hook(*args, **kwargs)
                return r

        else:
            # Wrapper that executes the decorated function after the class method
            def hooked(*args, **kwargs):
                r = func_to_hook(*args, **kwargs)
                func(*args, **kwargs)
                return r

        # Set the wrapper function as the hooked function
        _set_hooked_func(cls, hook, hooked)

    return decorator



import abc
import csv
import io
import typing

import psutil
from tqdm import tqdm

from .. import config
from ..config import refs
from ..exceptions import ConfigurationError
from ..mixins import NotParallel
from ..storage.interfaces import PlacementSet
from .strategy import ConnectionStrategy

if typing.TYPE_CHECKING:
    from ..cell_types import CellType
    from ..storage._files import FileDependencyNode
    from ..topology import Partition


@config.node
class ImportConnectivity(NotParallel, ConnectionStrategy, abc.ABC, classmap_entry=None):
    source: "FileDependencyNode" = config.file(required=True)
    cell_types: list["CellType"] = config.reflist(refs.cell_type_ref, required=False)
    partitions: list["Partition"] = config.reflist(refs.partition_ref, required=False)

    @config.property(default=False)
    def cache(self):
        return self.source.cache

    @cache.setter
    def cache(self, value):
        self.source.cache = bool(value)

    def connect(self, pre, post):
        self.parse_source(pre, post)

    @abc.abstractmethod
    def parse_source(self, pre, post):
        pass


@config.node
class CsvImportConnectivity(ImportConnectivity):
    pre_header: str = config.attr(default="pre")
    post_header: str = config.attr(default="post")
    mapping_key: str = config.attr()
    delimiter: str = config.attr(default=",")
    progress_bar: bool = config.attr(type=bool, default=True)

    def __boot__(self):
        if (
            len(self.presynaptic.cell_types) != 1
            or len(self.postsynaptic.cell_types) != 1
        ):
            raise NotImplementedError(
                "CsvImportConnectivity is strictly from 1 cell type to 1 other cell type"
            )

    def parse_source(self, pre, post):
        pre = pre.placement[0]
        post = post.placement[0]
        if self.mapping_key:

            def make_maps(pre_chunks, post_chunks):
                return self._passover(pre, pre_chunks, post, post_chunks)

            passover_iter = self._load_balance(pre, post)
        else:

            def make_maps(pre_chunks, post_chunks):
                def flush(pre_block, post_block, other_block):
                    with pre.chunk_context(pre_chunks):
                        with post.chunk_context(post_chunks):
                            self.connect_cells(pre, post, pre_block, post_block)

                return lambda x: x, lambda x: x, flush

            passover_iter = ((pre.get_all_chunks(), post.get_all_chunks()),)
        if self.progress_bar:
            passover_iter = tqdm(
                passover_iter, desc="processed", unit="passes", total=len(passover_iter)
            )
        with self.source.provide_stream() as (fp, encoding):
            text = io.TextIOWrapper(fp, encoding=encoding, newline="")
            reader = csv.reader(text)
            headers = [h.strip().lower() for h in next(reader)]
            try:
                cols = [*map(headers.index, (self.pre_header, self.post_header))]
            except ValueError:
                raise ConfigurationError(
                    f"'{self.pre_header}' or '{self.post_header}' not found in "
                    f"'{self.source.file.uri}' column headers {headers}."
                )
            other_cols = [r for r in range(len(headers)) if r not in cols]
            self._other_colnames = [headers[c] for c in other_cols]
            for pre_chunks, post_chunks in passover_iter:
                mappers = make_maps(pre_chunks, post_chunks)
                flush = mappers[2]
                i = 0
                if self.progress_bar:
                    reader = tqdm(reader, desc="imported", unit=" lines")
                pre_block = []
                post_block = []
                other_block = []
                for line in reader:
                    ids = [map_(_safe_int(line[c])) for c, map_ in zip(cols, mappers)]
                    other = [_safe_float(line[c]) for c in other_cols]
                    if ids[0] > -1 and ids[1] > -1:
                        pre_block.append(ids[0])
                        post_block.append(ids[1])
                        other_block.append(other)
                    if i % 100 == 0:
                        est_memsize = (len(other_cols) + 2) * i * 24
                        av_mem = psutil.virtual_memory().available
                        if est_memsize > av_mem / 10:
                            flush(pre_block, post_block, other_block)
                            pre_block = []
                            post_block = []
                            other_block = []
                    i += 1
                flush(pre_block, post_block, other_block)

    def _passover(self, pre: PlacementSet, pre_chunks, post: PlacementSet, post_chunks):
        with pre.chunk_context(pre_chunks):
            data = pre.load_additional(self.mapping_key)
            pre_map = {m: i for i, m in enumerate(data)}.get
        with pre.chunk_context(pre_chunks):
            data = post.load_additional(self.mapping_key)
            post_map = {m: i for i, m in enumerate(data)}.get

        def flush(pre_block, post_block, other_block):
            with pre.chunk_context(pre_chunks):
                with post.chunk_context(post_chunks):
                    self.connect_cells(pre, post, pre_block, post_block)

        return pre_map, post_map, flush

    def _load_balance(self, pre, post, pre_chunks=None, post_chunks=None):
        meminfo = psutil.virtual_memory()
        # Div 16 for array copying safety measures
        if (len(pre) + len(post)) * 24 < meminfo.available / 16:
            return ((pre.get_all_chunks(), post.get_all_chunks()),)
        else:
            raise NotImplementedError(
                "Too many cells to fit into memory. "
                "Piecewise import not implemented yet. Open a GitHub issue."
            )


def _safe_float(value: typing.Any) -> float:
    try:
        return float(value)
    except ValueError:
        return float("nan")


def _safe_int(value: typing.Any) -> int:
    try:
        return int(float(value))
    except ValueError:
        return -1


__all__ = ["CsvImportConnectivity", "ImportConnectivity"]



import typing

import numpy as np

from .. import config
from ..config import types
from ..mixins import InvertedRoI
from .strategy import ConnectionStrategy

if typing.TYPE_CHECKING:
    from ..config import Distribution


@config.node
class Convergence(ConnectionStrategy):
    """
    Connect cells based on a convergence distribution, i.e. by connecting each source cell
    to X target cells.
    """

    convergence: "Distribution" = config.attr(type=types.distribution(), required=True)

    def connect(self):
        raise NotImplementedError("Needs to be restored, please open an issue.")


class AllToAll(ConnectionStrategy):
    """
    All to all connectivity between two neural populations
    """

    def connect(self, pre, post):
        for from_ps in pre.placement:
            fl = len(from_ps)
            for to_ps in post.placement:
                len_ = len(to_ps)
                ml = fl * len_
                src_locs = np.full((ml, 3), -1)
                dest_locs = np.full((ml, 3), -1)
                src_locs[:, 0] = np.repeat(np.arange(fl), len_)
                dest_locs[:, 0] = np.tile(np.arange(len_), fl)
                self.connect_cells(from_ps, to_ps, src_locs, dest_locs)


@config.node
class FixedIndegree(InvertedRoI, ConnectionStrategy):
    """
    Connect a group of postsynaptic cell types to ``indegree`` uniformly random
    presynaptic cells from all the presynaptic cell types.
    """

    indegree: int = config.attr(type=int, required=True)

    def connect(self, pre, post):
        in_ = self.indegree
        rng = np.random.default_rng()
        high = sum(len(ps) for ps in pre.placement)
        for ps in post.placement:
            l = len(ps)
            pre_targets = np.full((l * in_, 3), -1)
            post_targets = np.full((l * in_, 3), -1)
            ptr = 0
            for i in range(l):
                post_targets[ptr : ptr + in_, 0] = i
                pre_targets[ptr : ptr + in_, 0] = rng.choice(high, in_, replace=False)
                ptr += in_
            lowmux = 0
            for pre_ps in pre.placement:
                highmux = lowmux + len(pre_ps)
                demux_idx = (pre_targets[:, 0] >= lowmux) & (pre_targets[:, 0] < highmux)
                demuxed = pre_targets[demux_idx]
                demuxed[:, 0] -= lowmux
                self.connect_cells(pre_ps, ps, demuxed, post_targets[demux_idx])
                lowmux = highmux


__all__ = ["AllToAll", "Convergence", "FixedIndegree"]



"""
Service module. Register or access interfaces that may be provided, mocked or missing, but
should always behave neatly on import.
"""

from ._util import ErrorModule as _ErrorModule
from .mpi import MPIService as _MPIService
from .mpilock import MPILockModule as _MPILockModule

MPI = _MPIService()
"""
MPI service
"""
MPILock = _MPILockModule("mpilock")
"""
MPILock service
"""

from .pool import JobPool as _JobPool  # noqa
from .pool import WorkflowError

JobPool = _JobPool
"""
JobPool service
"""


def __getattr__(attr):
    return _ErrorModule(f"{attr} is not a registered service.")


def register_service(attr, provider):
    globals()[attr] = provider


__all__ = ["MPI", "MPILock", "JobPool", "register_service", "WorkflowError"]



# TODO: check for parallel support in the hdf5 provider, if it has it, provide noop

from ._util import MockModule


class MockedWindowController:
    def __init__(self, comm=None, master=0):
        from . import MPI

        if comm is None:
            comm = MPI
        self._comm = comm
        self._size = comm.get_size()
        self._rank = comm.get_rank()
        self._master = master
        self._mocked = True
        self._closed = False

    @property
    def master(self):
        return self._master

    @property
    def rank(self):
        return self._rank

    @property
    def closed(self):
        return self._closed

    def close(self):
        self._closed = True

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def read(self):
        return _NoopLock()

    def write(self):
        return _NoopLock()

    def single_write(self, handle=None, rank=None):
        if rank is None:
            rank = self._master
        fence = Fence(self._rank == rank)
        if self._rank == rank:
            return _NoopLock(handle=handle, fence=fence)
        elif handle:
            return _NoHandle()
        else:
            return fence


class _NoopLock:
    def __init__(self, handle=None, fence=None):
        self._locked = 0
        self._mocked = True
        self._handle = handle
        self._fence = fence

    def locked(self):
        return self._locked > 0

    def __enter__(self):
        self._locked += 1
        return self._acquire_lock()

    def _acquire_lock(self):
        if self._handle is not None:
            return self._handle
        elif self._fence is not None:
            return self._fence

    def __exit__(self, exc_type, exc_value, traceback):
        self._locked -= 1


class Fence:
    def __init__(self, access):
        self._access = access
        self._obj = None
        self._mocked = True

    def guard(self):
        if not self._access:
            raise FencedSignal()

    def share(self, obj):
        self._obj = obj

    def collect(self):
        return self._obj

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is FencedSignal:
            return True


class _NoHandle:
    def __init__(self):
        self._mocked = True

    def __enter__(self):
        return None

    def __exit__(self, exc_type, exc_value, traceback):
        pass


class FencedSignal(Exception):
    pass


class MPILockModule(MockModule):
    def sync(self, comm=None, master=0):
        return MockedWindowController(comm, master)

    def __call__(self, comm=None, master=None):
        return MockedWindowController(comm, master)



"""
This module contains the global options.

You can set options at the ``script`` level (which superceeds all other levels such as
environment variables or project settings).

.. code-block::

  import bsb.options
  from bsb import BsbOption

  class MyOption(BsbOption, cli=("my_setting",), env=("MY_SETTING",), script=("my_setting", "my_alias")):
      def get_default(self):
          return 4

  # Register the option into the `bsb.options` module
  MyOption.register()

  assert bsb.options.my_setting == 4
  bsb.options.my_alias = 6
  assert bsb.options.my_setting == 6

Your ``MyOption`` will also be available on all CLI commands as ``--my_setting`` and will
be read from the ``MY_SETTING`` environment variable.
"""

import functools

from ._options import ProfilingOption, VerbosityOption

# Store the module magic for unpolluted namespace copy
_module_magic = globals().copy()

import sys
import types

from . import option as _bsboptmod
from .exceptions import OptionError, ReadOnlyOptionError
from .plugins import discover
from .reporting import report

_options = {}
_project_options = {}
_module_options = {}
_module_option_values = {}

# Everything defined between pre-freeze and post-freeze may be considered to be added to
# the `_OptionsModule` instance, place all public API functions between them.
_pre_freeze = set(globals().keys())


def _get_module_option(tag):  # pragma: nocover
    global _module_options

    if tag not in _module_options:
        if not discover_options.cache_info().misses:
            discover_options()
            return _get_module_tag(tag)
        else:
            raise OptionError(f"Unknown module option '{tag}'")
    return _module_options[tag]


def _get_module_tag(tag):  # pragma: nocover
    return _get_module_option(tag).__class__.script.tags[0]


def get_option_classes():
    """
    Return all of the classes that are used to create singleton options from. Useful to
    access the option descriptors rather than the option values.

    :returns: The classes of all the installed options by name.
    :rtype: dict[str, bsb.option.BsbOption]
    """
    return discover("options")


def get_option_descriptor(name):
    """
    Return an option

    :param name: Name of the option to look for.
    :type name: str
    :returns: The option singleton of that name.
    :rtype: dict[str, bsb.option.BsbOption]
    """
    global _options

    discover_options()
    if name in _options:
        return _options[name]
    else:
        raise OptionError(f"Unknown option '{name}'")


def register_option(name, option):
    """
    Register an option as a global BSB option. Options that are installed by the plugin
    system are automatically registered on import of the BSB.

    :param name: Name for the option, used to store and retrieve its singleton.
    :type name: str
    :param option: Option instance, to be used as a singleton.
    :type option: :class:`.option.BsbOption`
    """
    global _options

    if name in _options:
        if type(_options[name]) != type(option):
            raise OptionError(
                f"The '{name}' option name is already taken by {_options[name].__class__}."
            )
    else:
        _options[name] = option

        for tag in type(option).script.tags:
            _register_module_option(tag, option)
        if type(option).project.tags:
            _register_project_option(option)


def unregister_option(option):
    """
    Unregister a globally registered option. Also removes its script and project parts.

    :param option: Option singleton, to be removed.
    :type option: :class:`.option.BsbOption`
    """
    global _options, _project_options

    del _options[option.name]
    _remove_module_tags(*type(option).script.tags)
    path = type(option).project.tags
    if path:
        section = _project_options
        for slug in path[:-1]:
            if slug in section:
                section = section[slug]
            else:
                return
        try:
            del section[path[-1]]
        except KeyError:
            pass


def _register_project_option(option):  # pragma: nocover
    """
    Register an option that can be manipulated from ``pyproject.toml``, unregistered
    options can be used, but :func:`.options.store` and :func:`.options.read` won't work.

    :param option: Option.
    :type option: :class:`.option.BsbOption`
    """
    global _project_options

    path = type(option).project.tags
    section = _project_options
    for slug in path[:-1]:
        section = section.setdefault(slug, {})

    if path[-1] in section:
        raise OptionError(
            f"The '{'.'.join(path)}' tag is already taken by {section[path[-1]].__class__}."
        )
    else:
        section[path[-1]] = option


def get_project_option(tag):
    """
    Find a project option

    :param tag: dot-separated path of the option. e.g. ``networks.config_link``.
    :type tag: str
    :returns: Project option instance
    :rtype: :class:`.option.BsbOption`
    """
    global _project_options
    discover_options()
    path = tag.split(".")
    section = _project_options
    for slug in path:
        if slug in section:
            section = section[slug]
        else:
            raise OptionError(f"The project option `{tag}` does not exist.")
    return section


def _register_module_option(tag, option):  # pragma: nocover
    """
    Register an option that can be manipulated from :mod:`bsb.options`.
    """
    global _module_options

    if tag in _module_options:
        raise OptionError(
            f"The '{tag}' tag is already taken by {_module_options[tag].__class__}."
        )
    else:
        _module_options[tag] = option


def _remove_module_tags(*tags):  # pragma: nocover
    """
    Removes tags.
    """
    global _module_options, _module_option_values
    for tag in tags:
        try:
            del _module_options[tag]
        except KeyError:
            pass
        try:
            del _module_option_values[tag]
        except KeyError:
            pass


def reset_module_option(tag):
    global _module_option_values
    opt = _get_module_option(tag)
    # Module option values always stored under the "module tag" (= tag 0)
    try:
        del _module_option_values[type(opt).script.tags[0]]
    except KeyError:
        pass


def set_module_option(tag, value):
    """
    Set the value of a module option. Does the same thing as ``setattr(options, tag,
    value)``.

    :param tag: Name the option is registered with in the module.
    :type tag: str
    :param value: New module value for the option
    :type value: Any
    """

    global _module_option_values

    option = _get_module_option(tag)
    if option.readonly:
        raise ReadOnlyOptionError("'%tag%' is a read-only option.", option, tag)
    mod_tag = _get_module_tag(tag)
    _module_option_values[mod_tag] = getattr(option, "setter", lambda x: x)(value)


def get_module_option(tag):
    """
    Get the value of a module option. Does the same thing as ``getattr(options, tag)``

    :param tag: Name the option is registered with in the module.
    :type tag: str
    """
    global _module_option_values, _module_options
    discover_options()
    tag = _get_module_tag(tag)

    if tag in _module_option_values:
        return _module_option_values[tag]
    else:
        return _module_options[tag].get()


def is_module_option_set(tag):
    """
    Check if a module option was set.

    :param tag: Name the option is registered with in the module.
    :type tag: str
    :returns: Whether the option was ever set from the module
    :rtype: bool
    """
    global _module_option_values

    return _get_module_tag(tag) in _module_option_values


def get_option_descriptors():
    """
    Get all the registered option singletons.
    """
    global _options

    discover_options()
    return _options.copy()


@functools.cache
def discover_options():
    # Register the discoverable options
    plugins = discover("options")
    for plugin in plugins.values():
        option = plugin()
        register_option(option.name, option)


def store_option(tag, value):
    """
    Store an option value permanently in the project settings.

    :param tag: Dot-separated path of the project option
    :type tag: str
    :param value: New value for the project option
    :type value: Any
    """
    get_project_option(tag).project = value


def read_option(tag=None):
    """
    Read an option value from the project settings. Returns all project settings if tag is
    omitted.

    :param tag: Dot-separated path of the project option
    :type tag: str
    :returns: Value for the project option
    :rtype: Any
    """
    if tag is None:
        path, content = _bsboptmod._pyproject_bsb()
        report(f"Reading project settings from '{path}'", level=4)
        return content
    else:
        return get_project_option(tag).get(prio="project")


def get_option(tag, prio=None):
    """
    Retrieve the cascaded value for an option.

    :param tag: Name the option is registered with.
    :type tag: str
    :param prio: Give priority to a type of value. Can be any of 'script', 'cli',
      'project', 'env'.
    :type prio: str
    :returns: (Possibly prioritized) value of the option.
    :rtype: Any
    """
    option = get_option_descriptor(tag)
    return option.get(prio=prio)


_post_freeze = set(globals().keys()).difference(_pre_freeze)


class _OptionsModule(types.ModuleType):
    __name__ = "bsb.options"

    def __getattr__(self, attr):
        if attr in ["__path__", "__warningregistry__", "__qualname__"]:
            # __path__:
            # Python uses `hasattr(module, '__path__')` to see if a module is a package
            # so we need to raise an AttributeError to make `hasattr` return False.
            # __warningregistry__:
            # The `unittest` module checks existence.
            # __qualname__:
            # Sphinx checks this
            raise super().__getattribute__(attr)
        return self.get_module_option(attr)

    def __setattr__(self, attr, value):
        self.set_module_option(attr, value)

    def __delattr__(self, attr):
        try:
            opt = _get_module_option(attr)
        except OptionError:
            raise super().__delattr__(attr) from None
        else:
            del opt.script


_om = _OptionsModule(__name__)
# Copy the module magic from the original module.
_om.__dict__.update(_module_magic)
# Copy over the intended API from the original module.
for _key, _value in zip(_post_freeze, map(globals().get, _post_freeze)):
    _om.__dict__[_key] = _value
# Set the module's public API.
_om.__dict__["__all__"] = sorted([k for k in vars(_om).keys() if not k.startswith("_")])

sys.modules[__name__] = _om

register_option("verbosity", VerbosityOption())
register_option("profiling", ProfilingOption())

# Static public API
__all__ = [
    "get_option",
    "get_module_option",
    "get_option_descriptor",
    "get_option_classes",
    "get_option_descriptors",
    "get_project_option",
    "is_module_option_set",
    "read_option",
    "register_option",
    "reset_module_option",
    "set_module_option",
    "store_option",
    "unregister_option",
]



import typing
from os import PathLike

from .parser import MorphologyParser


def parse_morphology_content(content: typing.Union[str, bytes], parser="bsb", **kwargs):
    return MorphologyParser(parser=parser, **kwargs).parse_content(content)


def parse_morphology_file(file: typing.Union[str, PathLike], parser="bsb", **kwargs):
    return MorphologyParser(parser=parser, **kwargs).parse(file)


__all__ = ["parse_morphology_content", "parse_morphology_file"]



import math

import numpy as np

from .. import config
from ..config import types
from ..mixins import NotParallel
from ..reporting import report
from .strategy import PlacementStrategy


@config.node
class ParallelArrayPlacement(NotParallel, PlacementStrategy):
    """
    Implementation of the placement of cells in parallel arrays.
    """

    spacing_x: float = config.attr(type=float, required=True)
    angle: float = config.attr(type=types.deg_to_radian(), required=True)

    def place(self, chunk, indicators):
        """
        Cell placement: Create a lattice of parallel arrays/lines in the layer's surface.
        """
        for indicator in indicators.values():
            cell_type = indicator.cell_type
            radius = indicator.get_radius()
            for prt in self.partitions:
                width, depth, height = prt.data.mdc - prt.data.ldc
                ldc = prt.data.ldc
                # Extension of a single array in the X dimension
                spacing_x = self.spacing_x
                # Add a random shift to the starting points of the arrays for variation.
                x_shift = np.random.rand() * spacing_x
                # Place purkinje cells equally spaced over the entire length of the X axis kept apart by their dendritic trees.
                # They are placed in straight lines, tilted by a certain angle by adding a shifting value.
                x_pos = np.arange(start=0.0, stop=width, step=spacing_x) + x_shift
                if x_pos.shape[0] == 0:
                    # When the spacing_x of is larger than the simulation volume,
                    # place a single row on a random position along the x axis
                    x_pos = np.array([x_shift])
                # Amount of parallel arrays of cells
                n_arrays = x_pos.shape[0]
                # Number of cells
                n = np.sum(indicator.guess(prt.data))
                # Add extra cells to fill the lattice error volume which will be pruned
                n += int((n_arrays * spacing_x % width) / width * n)
                # cells to distribute along the rows
                cells_per_row = round(n / n_arrays)
                # The rounded amount of cells that will be placed
                cells_placed = cells_per_row * n_arrays
                # Calculate the position of the cells along the z-axis.
                y_pos, y_axis_distance = np.linspace(
                    start=0.0,
                    stop=depth - radius,
                    num=cells_per_row,
                    retstep=True,
                    endpoint=False,
                )
                # Center the cell soma center to the middle of the unit cell
                y_pos += radius + y_axis_distance / 2
                # The length of the X axis rounded up to a multiple of the unit cell size.
                lattice_x = n_arrays * spacing_x
                # The length of the X axis where cells can be placed in.
                bounded_x = lattice_x - radius * 2
                # Epsilon: open space in the unit cell along the z-axis
                epsilon = y_axis_distance - radius * 2
                # Storage array for the cells
                cells = np.empty((cells_placed, 3))
                for i in range(y_pos.shape[0]):
                    # Shift the arrays at an angle
                    angleShift = y_pos[i] * math.tan(self.angle)
                    # Apply shift and offset
                    x = x_pos + angleShift
                    # Place the cells in a bounded lattice with a little modulus magic
                    x = ldc[0] + x % bounded_x + radius
                    # Place the cells in their y-position with jitter
                    y = ldc[1] + y_pos[i] + epsilon * (np.random.rand(x.shape[0]) - 0.5)
                    # Place them at a uniformly random height throughout the partition.
                    z = ldc[2] + np.random.uniform(radius, height - radius, x.shape[0])
                    # Store this stack's cells
                    cells[(i * len(x)) : ((i + 1) * len(x)), 0] = x
                    cells[(i * len(x)) : ((i + 1) * len(x)), 1] = y
                    cells[(i * len(x)) : ((i + 1) * len(x)), 2] = z
                # Place all the cells in 1 batch (more efficient)
                positions = cells[cells[:, 0] < width - radius]

                # Determine in which chunks the cells must be placed
                cs = self.scaffold.configuration.network.chunk_size
                chunks_list = np.array(
                    [chunk.data + np.floor_divide(p, cs[0]) for p in positions]
                )
                unique_chunks_list = np.unique(chunks_list, axis=0)

                # For each chunk, place the cells
                for c in unique_chunks_list:
                    idx = np.where((chunks_list == c).all(axis=1))
                    pos_current_chunk = positions[idx]
                    self.place_cells(indicator, pos_current_chunk, chunk=c)
                report(f"Placed {len(positions)} {cell_type.name} in {prt.name}", level=3)


__all__ = ["ParallelArrayPlacement"]



import typing

import numpy as np

from ..exceptions import ChunkError

_iinfo = np.iinfo(np.int16)
Chunklike = typing.Union["Chunk", "numpy.typing.ArrayLike"]


class Chunk(np.ndarray):
    """
    Chunk identifier, consisting of chunk coordinates and size.
    """

    def __new__(cls, chunk, chunk_size):
        if any(c < _iinfo.min or c > _iinfo.max for c in chunk):
            raise ChunkError(
                f"Chunk coordinates must be between {_iinfo.min} and {_iinfo.max}."
            )
        obj = super().__new__(cls, (3,), dtype=np.short)
        obj[:] = chunk
        obj._size = np.array(chunk_size, dtype=float)
        return obj

    def __array_finalize__(self, obj):
        if obj is not None:
            self._size = getattr(obj, "_size", None)

    def __ne__(self, other):
        return self.id != other.id

    def __eq__(self, other):
        self_id, other_id = _safe_ids(self, other)
        return self_id == other_id

    def __gt__(self, other):
        self_id, other_id = _safe_ids(self, other)
        return self_id > other_id

    def __lt__(self, other):
        self_id, other_id = _safe_ids(self, other)
        return self_id < other_id

    def __ge__(self, other):
        self_id, other_id = _safe_ids(self, other)
        return self_id >= other_id

    def __le__(self, other):
        self_id, other_id = _safe_ids(self, other)
        return self_id <= other_id

    def __hash__(self):
        return int(self.id)

    def __reduce__(self):
        # Pickle ourselves, appending the `_size` attribute to our reduced state
        pickled_state = super().__reduce__()
        new_state = pickled_state[2] + (self._size,)
        return (pickled_state[0], pickled_state[1], new_state)

    def __setstate__(self, state):
        # Unpickle ourselves, grabbing the state we appended for `_size`
        super().__setstate__(state[:-1])
        self._size = state[-1]

    def _safe_id(self):
        return int(self) if self.shape == () else self.id

    @property
    def dimensions(self):
        return self._size

    @dimensions.setter
    def dimensions(self, value):
        self._size = np.array(value, dtype=float)

    @property
    def id(self):
        return sum(int(n) * 2 ** (i * 16) for i, n in enumerate(self.astype(np.uint16)))

    @property
    def box(self):
        return np.array(np.concatenate((self.ldc, self.mdc)), copy=False)

    @property
    def ldc(self):
        return np.array(self._size * self.astype(np.float64), copy=False)

    @property
    def mdc(self):
        return np.array(self._size * (self.astype(np.float64) + 1), copy=False)

    @classmethod
    def from_id(cls, id, size):
        return cls(
            np.uint16(
                [id % 2**16, (id // 2**16) % 2**16, (id // 2**32) % 2**16]
            ).astype(np.int16),
            size,
        )


def chunklist(chunks) -> typing.List[Chunk]:
    """
    Convert an iterable of chunk like objects to a sorted unique chunklist
    """
    return sorted(set(c if isinstance(c, Chunk) else Chunk(c, None) for c in chunks))


def _safe_ids(self, other):
    return (
        np.array(self, copy=False).view(Chunk)._safe_id(),
        np.array(other, copy=False).view(Chunk)._safe_id(),
    )


__all__ = ["Chunk", "chunklist"]



import atexit
import cProfile
import functools
import pickle
import sys
import traceback
import warnings
from functools import cache
from time import time
from uuid import uuid4

from .services import MPI


class Meter:
    def __init__(self, name, data):
        self.name = name
        self.data = data
        self._starts = []
        self._stops = []

    def __enter__(self):
        self.start()

    def start(self):
        self._starts.append(time())

    def __exit__(self, exc, exc_type, tb):
        self.stop()
        self._exc = exc, exc_type, traceback.format_tb(tb)

    def stop(self):
        self._stops.append(time())

    @property
    def calls(self):
        return len(self._starts)

    @property
    def elapsed(self):
        return sum(e - s for s, e in zip(self._starts, self._stops))


class ProfilingSession:
    def __init__(self):
        self._started = False
        self._meters = []
        self.name = "bsb_profiling"
        self._current_f = None
        self._flushcounter = 0

    def set_name(self, name):
        self.name = name

    @property
    def meters(self):
        return self._meters.copy()

    def start(self):
        if not self._started:
            self._started = True
            self.profile = cProfile.Profile()
            self.profile.enable()
            atexit.register(self.flush)

    def stop(self):
        if self._started:
            self._started = False
            self.profile.disable()
            atexit.unregister(self.flush)

    def meter(self, name, **data):
        meter = Meter(name, data)
        self._meters.append(meter)
        return meter

    def flush(self, stats=True):
        profile = self.profile
        if self._current_f is None:
            uuid = uuid4()
            self._current_f = f"{self.name}_{MPI.get_rank()}_{uuid}"
        if stats:
            self.profile.dump_stats(f"{self._current_f}_{self._flushcounter}.prf")
            self._flushcounter += 1
        try:
            del self.profile
            with open(f"{self._current_f}.pkl", "wb") as f:
                pickle.dump(self, f)
        except Exception as e:
            warnings.warn(f"Could not store profile: {e}")
        finally:
            self.profile = profile

    def view(self):
        try:
            from snakeviz.cli import main as snakeviz
        except ImportError:
            raise ImportError("Please `pip install snakeviz` to view profiles.") from None

        args = sys.argv
        if self._current_f is None:
            self.flush()
        sys.argv = ["snakeviz", f"{self._current_f}.prf"]

        snakeviz()

        sys.argv = args

    @staticmethod
    def load(fstem):
        with open(f"{fstem}.pkl", "rb") as f:
            return pickle.load(f)


@cache
def get_active_session():
    return ProfilingSession()


def activate_session(name=None):
    session = get_active_session()
    if name is not None:
        session.set_name(name)
    session.start()
    return session


def node_meter(*methods):
    def get_node_method_name(method, args, kwargs):
        return (
            f"{args[0].get_node_name()}[{args[0].__class__.__name__}].{method.__name__}"
        )

    def decorator(node_cls):
        for method_name in methods:
            if method := getattr(node_cls, method_name, None):
                setattr(node_cls, method_name, meter(method, name_f=get_node_method_name))

        return node_cls

    return decorator


def meter(f=None, *, name_f=None):
    def decorated(*args, **kwargs):
        import bsb.options

        if bsb.options.profiling:
            session = get_active_session()
            if name_f:
                name = name_f(f, args, kwargs)
            else:
                name = f.__name__
            with session.meter(name, args=str(args), kwargs=str(kwargs)):
                r = f(*args, **kwargs)
            session.flush(stats=False)
            return r
        else:
            return f(*args, **kwargs)

    if f is None:

        def decorator(g):
            nonlocal f
            f = g
            return functools.wraps(f)(decorated)

        return decorator
    else:
        return functools.wraps(f)(decorated)


def view_profile(fstem):
    ProfilingSession.load(fstem).view()


__all__ = [
    "Meter",
    "ProfilingSession",
    "activate_session",
    "get_active_session",
    "meter",
    "node_meter",
    "view_profile",
]



import abc
import itertools
import typing

import numpy as np

from .. import config
from .._util import obj_str_insert
from ..config import refs, types
from ..config._attrs import cfgdict
from ..exceptions import DistributorError, EmptySelectionError
from ..mixins import HasDependencies
from ..profiling import node_meter
from ..reporting import warn
from ..storage._chunks import Chunk
from ..voxels import VoxelSet
from .distributor import DistributorsNode
from .indicator import PlacementIndications, PlacementIndicator

if typing.TYPE_CHECKING:
    from ..cell_types import CellType
    from ..core import Scaffold
    from ..topology import Partition


@config.dynamic(attr_name="strategy", required=True)
class PlacementStrategy(abc.ABC, HasDependencies):
    """
    Quintessential interface of the placement module. Each placement strategy defines an
    approach to placing neurons into a volume.
    """

    scaffold: "Scaffold"

    name: str = config.attr(key=True)
    cell_types: list["CellType"] = config.reflist(refs.cell_type_ref, required=True)
    partitions: list["Partition"] = config.reflist(refs.partition_ref, required=True)
    overrides: cfgdict["PlacementIndications"] = config.dict(type=PlacementIndications)
    depends_on: list["PlacementStrategy"] = config.reflist(refs.placement_ref)
    distribute: DistributorsNode = config.attr(
        type=DistributorsNode, default=dict, call_default=True
    )
    indicator_class = PlacementIndicator

    def __init_subclass__(cls, **kwargs):
        super(cls, cls).__init_subclass__(**kwargs)
        # Decorate subclasses to measure performance
        node_meter("place")(cls)

    def __hash__(self):
        return id(self)

    def __lt__(self, other):
        # This comparison should sort placement strategies by name, via __repr__ below
        return str(self) < str(other)

    @obj_str_insert
    def __repr__(self):
        config_name = self.name
        if not hasattr(self, "scaffold"):
            return f"'{config_name}'"
        part_str = ""
        if len(self.partitions):
            partition_names = [p.name for p in self.partitions]
            part_str = f" into {partition_names}"
        ct_names = [ct.name for ct in self.cell_types]
        return f"'{config_name}', placing {ct_names}{part_str}"

    @abc.abstractmethod
    def place(self, chunk, indicators):
        """
        Central method of each placement strategy. Given a chunk, should fill that chunk
        with cells by calling the scaffold's (available as ``self.scaffold``)
        :func:`~bsb.core.Scaffold.place_cells` method.

        :param chunk: Chunk to fill
        :type chunk: bsb.storage._chunks.Chunk
        :param indicators: Dictionary of each cell type to its PlacementIndicator
        :type indicators: Mapping[str, bsb.placement.indicator.PlacementIndicator]
        """
        pass

    def place_cells(self, indicator, positions, chunk, additional=None):
        if additional is None:
            additional = {}
        if self.distribute._has_mdistr() or indicator.use_morphologies():
            selector_error = None
            try:
                morphologies, rotations = self.distribute._specials(
                    self.partitions, indicator, positions
                )
            except EmptySelectionError as e:
                selector_error = ", ".join(str(s) for s in e.selectors)
            if selector_error:
                # Starting from Python 3.11, even though we raise from None, the original
                # EmptySelectionError somehow still gets pickled and contains unpicklable
                # elements. So we work around by raising here, outside of the exception
                # context.
                raise DistributorError(
                    "Morphology distribution couldn't find any"
                    + f" morphologies with the following selector(s): {selector_error}"
                )
        elif self.distribute._has_rdistr():
            rotations = self.distribute(
                "rotations", self.partitions, indicator, positions
            )
            morphologies = None
        else:
            morphologies, rotations = None, None

        distr = self.distribute._curry(self.partitions, indicator, positions)
        additional.update(
            {prop: distr(prop) for prop in self.distribute.properties.keys()}
        )
        self.scaffold.place_cells(
            indicator.cell_type,
            positions=positions,
            rotations=rotations,
            morphologies=morphologies,
            additional=additional,
            chunk=chunk,
        )

    def queue(self, pool, chunk_size):
        """
        Specifies how to queue this placement strategy into a job pool. Can be overridden,
        the default implementation asks each partition to chunk itself and creates 1
        placement job per chunk.
        """
        # Get the queued jobs of all the strategies we depend on.
        deps = set(
            itertools.chain(
                *(pool.get_submissions_of(strat) for strat in self.get_deps())
            )
        )
        for p in self.partitions:
            chunks = p.to_chunks(chunk_size)
            for chunk in chunks:
                job = pool.queue_placement(self, Chunk(chunk, chunk_size), deps=deps)

    def is_entities(self):
        return "entities" in self.__class__.__dict__ and self.__class__.entities

    def get_indicators(self):
        """
        Return indicators per cell type. Indicators collect all configuration information
        into objects that can produce guesses as to how many cells of a type should be
        placed in a volume.
        """
        return {
            ct.name: self.__class__.indicator_class(self, ct) for ct in self.cell_types
        }

    def guess_cell_count(self):
        return sum(ind.guess() for ind in self.get_indicators().values())

    def get_deps(self):
        return set(self.depends_on)


@config.node
class FixedPositions(PlacementStrategy):
    positions: np.ndarray = config.attr(type=types.ndarray())

    def place(self, chunk, indicators):
        if self.positions is None:
            raise ValueError(
                f"Please set `.positions` on '{self.name}' before placement."
            )
        if not len(self.positions):
            warn(f"No positions given to {self.get_node_name()}.")
            return
        for indicator in indicators.values():
            inside_chunk = VoxelSet([chunk], chunk.dimensions).inside(self.positions)
            self.place_cells(indicator, self.positions[inside_chunk], chunk)

    def guess_cell_count(self):
        if self.positions is None:
            raise ValueError(f"Please set `.positions` on '{self.name}'.")
        return len(self.positions)

    def queue(self, pool, chunk_size):
        if self.positions is None:
            raise ValueError(f"Please set `.positions` on '{self.name}'.")
        # Get the queued jobs of all the strategies we depend on.
        deps = set(
            itertools.chain(
                *(pool.get_submissions_of(strat) for strat in self.get_deps())
            )
        )
        for chunk in VoxelSet.fill(self.positions, chunk_size):
            pool.queue_placement(self, Chunk(chunk, chunk_size), deps=deps)


@config.node
class Entities(PlacementStrategy):
    """
    Implementation of the placement of entities that do not have a 3D position,
    but that need to be connected with other cells of the network.
    """

    def queue(self, pool, chunk_size):
        # Entities ignore chunks since they don't intrinsically store any data.
        pool.queue_placement(self, Chunk([0, 0, 0], chunk_size))

    def place(self, chunk, indicators):
        for indicator in indicators.values():
            cell_type = indicator.cell_type
            # Guess total number, not chunk number, as entities bypass chunking.
            n = sum(
                # Pass the voxelset if it exists
                np.sum(indicator.guess(voxels=getattr(p, "voxelset", None)))
                for p in self.partitions
            )
            self.scaffold.create_entities(cell_type, n)


__all__ = ["Entities", "FixedPositions", "PlacementStrategy"]



import abc
import typing
from itertools import chain

from .. import config
from .._util import ichain, obj_str_insert
from ..config import refs, types
from ..exceptions import ConnectivityError
from ..mixins import HasDependencies
from ..profiling import node_meter
from ..reporting import warn

if typing.TYPE_CHECKING:
    from ..cell_types import CellType
    from ..core import Scaffold
    from ..morphologies import MorphologySet
    from ..services import JobPool
    from ..storage.interfaces import PlacementSet


@config.node
class Hemitype:
    """
    Class used to represent one (pre- or postsynaptic) side of a connection rule.
    """

    scaffold: "Scaffold"

    cell_types: list["CellType"] = config.reflist(refs.cell_type_ref, required=True)
    """List of cell types to use in connection."""
    labels: list[str] = config.attr(type=types.list())
    """List of labels to filter the placement set by."""
    morphology_labels: list[str] = config.attr(type=types.list())
    """List of labels to filter the morphologies by."""
    morpho_loader: typing.Callable[["PlacementSet"], "MorphologySet"] = config.attr(
        type=types.function_(),
        required=False,
        call_default=False,
        default=(lambda ps: ps.load_morphologies()),
    )
    """
    Function to load the morphologies (MorphologySet) from a PlacementSet. This override
    can allow temporary dynamic morphology generation during the connectivity phase, from
    a much smaller, or empty, MorphologySet. It is useful for example when the task would
    take too much disk space or time otherwise.
    """


class HemitypeCollection:
    def __init__(self, hemitype, roi):
        self.hemitype = hemitype
        self.roi = roi

    def __iter__(self):
        return iter(self.hemitype.cell_types)

    @property
    def placement(self):
        return [
            ct.get_placement_set(
                chunks=self.roi,
                labels=self.hemitype.labels,
                morphology_labels=self.hemitype.morphology_labels,
            )
            for ct in self.hemitype.cell_types
        ]


@config.dynamic(attr_name="strategy", required=True)
class ConnectionStrategy(abc.ABC, HasDependencies):
    scaffold: "Scaffold"
    name: str = config.attr(key=True)
    """Name used to refer to the connectivity strategy"""
    presynaptic: Hemitype = config.attr(type=Hemitype, required=True)
    """Presynaptic (source) neuron population"""
    postsynaptic: Hemitype = config.attr(type=Hemitype, required=True)
    """Postsynaptic (target) neuron population"""
    depends_on: list["ConnectionStrategy"] = config.reflist(refs.connectivity_ref)
    """The list of strategies that must run before this one"""
    output_naming: typing.Union[str, None, dict[str, dict[str, str, None, list[str]]]] = (
        config.attr(
            type=types.or_(
                types.str(),
                types.dict(
                    type=types.dict(
                        type=types.or_(
                            types.str(), types.list(type=types.str()), types.none()
                        )
                    )
                ),
                types.list(type=types.str()),
            )
        )
    )
    """Specifies how to name the output ConnectivitySets in which the connections between cell type pairs are stored."""

    def __init_subclass__(cls, **kwargs):
        super(cls, cls).__init_subclass__(**kwargs)
        # Decorate subclasses to measure performance
        node_meter("connect")(cls)

    def __hash__(self):
        return id(self)

    def __lt__(self, other):
        # This comparison should sort connection strategies by name, via __repr__ below
        return str(self) < str(other)

    @obj_str_insert
    def __repr__(self):
        if not hasattr(self, "scaffold"):
            return f"'{self.name}'"
        pre = [ct.name for ct in self.presynaptic.cell_types]
        post = [ct.name for ct in self.postsynaptic.cell_types]
        return f"'{self.name}', connecting {pre} to {post}"

    @abc.abstractmethod
    def connect(self, presyn_collection, postsyn_collection):
        pass

    def get_deps(self):
        return set(self.depends_on)

    def _get_connect_args_from_job(self, pre_roi, post_roi):
        pre = HemitypeCollection(self.presynaptic, pre_roi)
        post = HemitypeCollection(self.postsynaptic, post_roi)
        return pre, post

    def connect_cells(self, pre_set, post_set, src_locs, dest_locs, tag=None):
        names = self.get_output_names(pre_set.cell_type, post_set.cell_type)
        between_msg = f"between {pre_set.cell_type.name} and {post_set.cell_type.name}"
        if len(names) == 0:
            raise ConnectivityError(
                f"Connections {between_msg} have been disabled by output naming."
            )
        elif len(names) == 1:
            name = names[0]
            if tag is not None and tag != name:
                raise ConnectivityError(
                    f"Tag ('{tag}') and output name ('{name}') mismatch."
                )
        else:
            names_msg = f"{between_msg} (names: {', '.join(names)})."
            if tag is None:
                raise ConnectivityError(
                    f"No tag was given to decide between multiple output names {names_msg}"
                )
            elif tag not in names:
                raise ConnectivityError(
                    f"Tag '{tag}' is not a valid output name {names_msg}"
                )
            else:
                name = tag

        cs = self.scaffold.require_connectivity_set(
            pre_set.cell_type, post_set.cell_type, name
        )
        cs.connect(pre_set, post_set, src_locs, dest_locs)

    def get_region_of_interest(self, chunk):
        pass

    def queue(self, pool: "JobPool"):
        """
        Specifies how to queue this connectivity strategy into a job pool. Can
        be overridden, the default implementation asks each partition to chunk
        itself and creates 1 placement job per chunk.
        """
        # Get the queued jobs of all the strategies we depend on.
        dep_jobs = set(
            chain.from_iterable(
                pool.get_submissions_of(strat) for strat in self.get_deps()
            )
        )
        pre_types = self.presynaptic.cell_types
        # Iterate over each chunk that is populated by our presynaptic cell types.
        from_chunks = set(
            chain.from_iterable(
                ct.get_placement_set().get_all_chunks() for ct in pre_types
            )
        )
        rois = {
            chunk: roi
            for chunk in from_chunks
            if (roi := self.get_region_of_interest(chunk)) is None or len(roi)
        }
        if not rois:
            warn(
                f"No overlap found between {[pre.name for pre in pre_types]} and "
                f"{[post.name for post in self.postsynaptic.cell_types]} "
                f"in '{self.name}'."
            )
        for chunk, roi in rois.items():
            job = pool.queue_connectivity(self, [chunk], roi, deps=dep_jobs)

    def get_cell_types(self):
        return set(self.presynaptic.cell_types) | set(self.postsynaptic.cell_types)

    def get_all_pre_chunks(self):
        all_ps = (ct.get_placement_set() for ct in self.presynaptic.cell_types)
        chunks = set(ichain(ps.get_all_chunks() for ps in all_ps))
        return list(chunks)

    def get_all_post_chunks(self):
        all_ps = (ct.get_placement_set() for ct in self.postsynaptic.cell_types)
        chunks = set(ichain(ps.get_all_chunks() for ps in all_ps))
        return list(chunks)

    def get_output_names(self, pre=None, post=None):
        if (pre is None) != (post is None):
            raise RuntimeError("pre and post must be specified or omitted together.")
        if pre is not None and (
            pre not in self.presynaptic.cell_types
            or post not in self.postsynaptic.cell_types
        ):
            raise ValueError(
                f"'{pre.name}' and '{post.name}' are not a valid cell pair type for this connectivity strategy."
            )
        if self.output_naming is None or isinstance(self.output_naming, str):
            return self._infer_output_name(self.output_naming or self.name, pre, post)
        elif isinstance(self.output_naming, list):
            # Call `_infer_output_name` for each given `base` in the list, and chain them together
            return [
                *ichain(
                    self._infer_output_name(base, pre, post)
                    for base in self.output_naming
                )
            ]
        else:
            return self._get_output_name(pre, post)

    def _infer_output_name(self, base, pre, post):
        if len(self.presynaptic.cell_types) > 1 or len(self.postsynaptic.cell_types) > 1:
            if pre is None:
                # All output names
                return [
                    *ichain(
                        self._infer_output_name(base, pre_ct, post_ct)
                        for pre_ct in self.presynaptic.cell_types
                        for post_ct in self.postsynaptic.cell_types
                    )
                ]
            else:
                # Pair specific output name
                return [f"{base}_{pre.name}_to_{post.name}"]
        else:
            # Single output name
            return [base]

    def _get_output_name(self, pre, post):
        if pre is None:
            # All output names
            return [
                *ichain(
                    self._get_output_name(pre_ct, post_ct)
                    for pre_ct in self.presynaptic.cell_types
                    for post_ct in self.postsynaptic.cell_types
                )
            ]
        else:
            # Pair specific output name
            MISSING = type("MISSING", (), {"get": lambda *args: MISSING})()
            spec = self.output_naming.get(pre.name, MISSING).get(post.name, MISSING)
            if spec is MISSING:
                return self._infer_output_name(self.name, pre, post)
            elif spec is None:
                return []
            elif isinstance(spec, str):
                return [spec]
            else:
                return spec


__all__ = ["ConnectionStrategy", "Hemitype", "HemitypeCollection"]



import itertools
import random

import numpy as np
from numpy.random import default_rng

from ... import config
from ..._util import ichain
from ...config import types
from ..strategy import ConnectionStrategy
from .shared import Intersectional

_rng = default_rng()


@config.node
class VoxelIntersection(Intersectional, ConnectionStrategy):
    """
    This strategy finds overlap between voxelized morphologies.
    """

    contacts = config.attr(type=types.distribution(), default=1)
    voxels_pre = config.attr(type=int, default=50)
    voxels_post = config.attr(type=int, default=50)
    cache = config.attr(type=bool, default=True)
    favor_cache = config.attr(type=types.in_(["pre", "post"]), default="pre")

    def connect(self, pre, post):
        # Note on the caching terms: `targets` are the population that will be cached the
        # strongest; their voxelized tree will remain in place, while the candidates are
        # rotated and translated to overlap the target tree.
        # The choice to make something cached harder is if they have less different
        # morphologies, and good choices for candidates are the population with more
        # numerous and smaller morphologies.
        if self.favor_cache == "pre":
            targets = pre
            candidates = post
            self._n_tvoxels = self.voxels_pre
            self._n_cvoxels = self.voxels_post
            target_morpho = self.presynaptic.morpho_loader
            cand_morpho = self.postsynaptic.morpho_loader
        else:
            targets = post
            candidates = pre
            self._n_tvoxels = self.voxels_post
            self._n_cvoxels = self.voxels_pre
            target_morpho = self.postsynaptic.morpho_loader
            cand_morpho = self.presynaptic.morpho_loader
        combo_itr = self.candidate_intersection(targets, candidates)
        mset_cache = {}
        for target_set, cand_set, match_itr in combo_itr:
            if self.cache:
                if id(target_set) not in mset_cache:
                    mset_cache[id(target_set)] = target_morpho(target_set)
                if id(cand_set) not in mset_cache:
                    mset_cache[id(cand_set)] = cand_morpho(cand_set)
                target_mset = mset_cache[id(target_set)]
                cand_mset = mset_cache[id(cand_set)]
            else:
                target_mset = target_morpho(target_set)
                cand_mset = cand_morpho(cand_set)
            self._match_voxel_intersection(
                match_itr, target_set, cand_set, target_mset, cand_mset
            )

    def _match_voxel_intersection(self, matches, tset, cset, tmset, cmset):
        # Soft-caching caches at the IO level and gives you a fresh copy of the morphology
        # each time, the `cached_voxelize` function we need wouldn't have any effect!
        tm_iter = tmset.iter_morphologies(cache=self.cache, hard_cache=self.cache)
        target_itrs = zip(tset.load_positions(), tset.load_rotations().iter(), tm_iter)
        rotations = cset.load_rotations()
        positions = cset.load_positions()
        data_acc = []
        for target, candidates in enumerate(matches):
            tpos, trot, tmor = next(target_itrs)
            if not len(candidates):
                # No need to load or voxelize if there's no candidates anyway
                continue
            # Load and voxelize the target into a box tree
            if self.cache:
                tvoxels = tmor.cached_voxelize(N=self._n_tvoxels)
            else:
                tvoxels = tmor.voxelize(N=self._n_tvoxels)
            tree = tvoxels.as_boxtree(cache=self.cache)
            for cand in candidates:
                cpos = positions[cand]
                crot = rotations[cand]
                # Don't hard cache, as we mutate the instance we get.
                morpho = cmset.get(cand, cache=self.cache, hard_cache=False)
                # Transform candidate, keep target unrotated and untranslated at origin:
                # 1) Rotate self by own rotation
                # 2) Translate by position relative to target
                # 3) Anti-rotate by target rotation
                # Gives us the candidate relative to the target without having to modify,
                # reload, recalculate or revoxelize any of the target morphologies.
                # So in the case of a single target morphology we can keep that around.
                morpho.rotate(crot)
                morpho.translate(cpos - tpos)
                morpho.rotate(trot.inv())
                cvoxels = morpho.voxelize(N=self._n_cvoxels)
                boxes = cvoxels.as_boxes()
                # Filter out the candidate voxels that overlap with target voxels.
                overlap = [(i, v) for i, v in enumerate(tree.query(boxes)) if v]
                if overlap:
                    locations = self._pick_locations(
                        target, cand, tvoxels, cvoxels, overlap
                    )
                    data_acc.append(locations)

        # Preallocating and filling is faster than `np.concatenate` :shrugs:
        acc_idx = np.cumsum(
            [len(a[0]) for a in data_acc],
        )
        # The inline if guards against the case where there's no overlap
        tlocs = np.empty((acc_idx[-1] if len(acc_idx) else 0, 3), dtype=int)
        clocs = np.empty((acc_idx[-1] if len(acc_idx) else 0, 3), dtype=int)
        for (s, e), (tblock, cblock) in zip(_pairs_with_zero(acc_idx), data_acc):
            tlocs[s:e] = tblock
            clocs[s:e] = cblock

        if self.favor_cache == "pre":
            src_set, dest_set = tset, cset
            src_locs, dest_locs = tlocs, clocs
        else:
            src_set, dest_set = cset, tset
            src_locs, dest_locs = clocs, tlocs

        self.connect_cells(src_set, dest_set, src_locs, dest_locs)

    def _pick_locations(self, tid, cid, tvoxels, cvoxels, overlap):
        n = int(self.contacts.draw(1)[0])
        if n <= 0:
            return np.empty((0, 3), dtype=int), np.empty((0, 3), dtype=int)
        cpool = cvoxels.get_data([c for c, _ in overlap])
        tpool = [tvoxels.get_data(t) for _, t in overlap]
        pool = np.column_stack(
            (
                np.repeat(cpool, [len(t) for t in tpool]),
                np.array([*ichain(tpool)], dtype=object),
            )
        )
        weights = [len(c) * len(t) for c, t in pool]
        tlocs = []
        clocs = []
        for cpick, tpick in random.choices(pool, weights, k=n):
            clocs.append((cid, *random.choice(cpick)))
            tlocs.append((tid, *random.choice(tpick)))
        return tlocs, clocs


def _pairs_with_zero(iterable):
    a, b = itertools.tee(iterable)
    try:
        yield 0, next(b)
    except StopIteration:
        pass
    else:
        yield from zip(a, b)


__all__ = ["VoxelIntersection"]



import builtins
import inspect
import sys

from .._contexts import get_cli_context, reset_cli_context
from ..exceptions import CommandError, DryrunError
from .commands import load_root_command


def handle_cli():
    handle_command(sys.argv[1:], exit=True)


def handle_command(command, dryrun=False, exit=False):
    reset_cli_context()
    context = get_cli_context()
    root_command = load_root_command()
    parser = root_command.get_parser(context)
    try:
        namespace = parser.parse_args(command)
    except CommandError as e:
        if exit:
            print(e)
            builtins.exit(1)
        else:
            raise
    if not dryrun:
        for action in namespace.internal_action_list or ():
            action(namespace)
    if not dryrun or _can_dryrun(namespace.handler, namespace):
        namespace.handler(namespace, dryrun=dryrun)
    else:  # pragma: nocover
        raise DryrunError(f"`{namespace.handler.__name__}` doesn't support dryruns.")
    return context


def _can_dryrun(handler, namespace):
    try:
        return bool(inspect.signature(handler).bind(namespace, dryrun=True))
    except TypeError:
        return False


__all__ = ["handle_cli", "handle_command"]



from dataclasses import dataclass as _dc

from .adapter import SimulatorAdapter as _SimAdapter
from .simulation import Simulation as _Sim


@_dc
class SimulationBackendPlugin:
    Adapter: _SimAdapter
    Simulation: _Sim


def get_simulation_adapter(name: str):
    from ._backends import get_simulation_adapters

    return get_simulation_adapters()[name]


__all__ = ["SimulationBackendPlugin", "get_simulation_adapter"]



"""
bsb.config module

Contains the dynamic attribute system; Use ``@bsb.config.root/node/dynamic/pluggable`` to
decorate your classes and add class attributes using ``x =
config.attr/dict/list/ref/reflist`` to populate your classes with powerful attributes.
"""

import builtins
import functools
import glob
import itertools
import os
import sys
import traceback
import typing
from shutil import copy2 as copy_file

from .. import plugins
from .._util import ichain
from ..exceptions import ConfigTemplateNotFoundError, ParserError
from . import refs, types
from ._attrs import (
    ConfigurationAttribute,
    attr,
    catch_all,
    dict,
    dynamic,
    file,
    list,
    node,
    pluggable,
    property,
    provide,
    ref,
    reflist,
    root,
    slot,
    unset,
)
from ._distributions import Distribution
from ._hooks import after, before, has_hook, on, run_hook
from ._make import (
    compose_nodes,
    get_config_attributes,
    walk_node_attributes,
    walk_nodes,
)
from .parsers import get_configuration_parser, get_configuration_parser_classes

if typing.TYPE_CHECKING:
    from ._config import Configuration


@functools.cache
def __getattr__(name):
    if name == "Configuration":
        # Load the Configuration class on demand, not on import, to avoid circular
        # dependencies.
        from ._config import Configuration

        return Configuration
    else:
        raise object.__getattribute__(sys.modules[__name__], name)


def get_config_path():
    import os

    env_paths = os.environ.get("BSB_CONFIG_PATH", None)
    if env_paths is None:
        env_paths = ()
    else:
        env_paths = env_paths.split(":")
    plugin_paths = plugins.discover("config.templates")
    return [*itertools.chain((os.getcwd(),), env_paths, *plugin_paths.values())]


def copy_configuration_template(template, output="network_configuration.json", path=None):
    path = [
        *map(
            os.path.abspath,
            itertools.chain(get_config_path(), path or ()),
        )
    ]
    for d in path:
        if files := glob.glob(os.path.join(d, template)):
            break
    else:
        raise ConfigTemplateNotFoundError(
            "'%template%' not found in config path %path%", template, path
        )
    copy_file(files[0], output)


def format_configuration_content(parser_name: str, config: "Configuration", **kwargs):
    """
    Convert a configuration object to a string using the given parser.
    """
    return get_configuration_parser(parser_name, **kwargs).generate(
        config.__tree__(), pretty=True
    )


def make_configuration_diagram(config):
    dot = f'digraph "{config.name or "network"}" {{'
    for c in config.cell_types.values():
        dot += f'\n  {c.name}[label="{c.name}"]'
    for name, conn in config.connectivity.items():
        for pre in conn.presynaptic.cell_types:
            for post in conn.postsynaptic.cell_types:
                dot += f'\n  {pre.name} -> {post.name}[label="{name}"];'
    dot += "\n}\n"
    return dot


def _try_parsers(content, classes, ext=None, path=None):  # pragma: nocover
    if ext is not None:

        def file_has_parser_ext(kv):
            return ext not in getattr(kv[1], "data_extensions", ())

        classes = builtins.dict(sorted(classes.items(), key=file_has_parser_ext))
    exc = {}
    for name, cls in classes.items():
        try:
            tree, meta = cls().parse(content, path=path)
        except Exception as e:
            if getattr(e, "_bsbparser_show_user", False):
                raise e from None
            exc[name] = e
        else:
            return (name, tree, meta)
    msges = [
        (
            f"- Can't parse contents with '{n}':\n",
            "".join(traceback.format_exception(type(e), e, e.__traceback__)),
        )
        for n, e in exc.items()
    ]
    if path:
        msg = f"Could not parse '{path}'"
    else:
        msg = f"Could not parse content string ({len(content)} characters long)"
    raise ParserError("\n".join(ichain(msges)) + f"\n{msg}")


def _from_parsed(parser_name, tree, meta, file=None):
    from ._config import Configuration

    conf = Configuration(tree)
    conf._parser = parser_name
    conf._meta = meta
    conf._file = file
    return conf


def parse_configuration_file(file, parser=None, path=None, **kwargs):
    if hasattr(file, "read"):
        data = file.read()
        try:
            path = str(path) or os.fspath(file)
        except TypeError:
            pass
    else:
        file = os.path.abspath(file)
        path = path or file
        with open(file, "r") as f:
            data = f.read()
    return parse_configuration_content(data, parser, path, **kwargs)


def parse_configuration_content(content, parser=None, path=None, **kwargs):
    if parser is None:
        parser_classes = get_configuration_parser_classes()
        ext = path.split(".")[-1] if path is not None else None
        parser_name, tree, meta = _try_parsers(content, parser_classes, ext, path=path)
    elif isinstance(parser, str):
        parser_name = parser
        parser = get_configuration_parser(parser_name, **kwargs)
        tree, meta = parser.parse(content, path=path)
    else:
        parser_name = parser.__name__
        tree, meta = parser.parse(content, path=path)
    return _from_parsed(parser_name, tree, meta, path)


# Static public API
__all__ = [
    "Configuration",
    "ConfigurationAttribute",
    "Distribution",
    "after",
    "attr",
    "before",
    "catch_all",
    "compose_nodes",
    "copy_configuration_template",
    "dict",
    "dynamic",
    "file",
    "format_configuration_content",
    "get_config_attributes",
    "get_config_path",
    "has_hook",
    "list",
    "make_configuration_diagram",
    "node",
    "on",
    "parse_configuration_file",
    "parse_configuration_content",
    "pluggable",
    "property",
    "provide",
    "ref",
    "refs",
    "reflist",
    "root",
    "run_hook",
    "slot",
    "types",
    "unset",
    "walk_node_attributes",
    "walk_nodes",
]
__api__ = [
    "Configuration",
    "ConfigurationAttribute",
    "Distribution",
    "compose_nodes",
    "copy_configuration_template",
    "format_configuration_content",
    "get_config_attributes",
    "get_config_path",
    "make_config_diagram",
    "parse_configuration_file",
    "parse_configuration_content",
    "refs",
    "types",
    "walk_node_attributes",
    "walk_nodes",
]



import typing
from time import time

from .. import config
from ..config import types as cfgtypes
from ..config._attrs import cfgdict, cfglist
from ._backends import get_simulation_nodes
from .cell import CellModel
from .connection import ConnectionModel
from .device import DeviceModel

if typing.TYPE_CHECKING:
    from ..cell_types import CellType
    from ..connectivity import ConnectionStrategy
    from ..core import Scaffold
    from ..storage.interfaces import ConnectivitySet


class ProgressEvent:
    def __init__(self, progression, duration, time):
        self.progression = progression
        self.duration = duration
        self.time = time


@config.pluggable(key="simulator", plugin_name="simulation backend")
class Simulation:
    scaffold: "Scaffold"
    simulator: str
    name: str = config.attr(key=True)
    duration: float = config.attr(type=float, required=True)
    cell_models: cfgdict[CellModel] = config.slot(type=CellModel, required=True)
    connection_models: cfgdict[ConnectionModel] = config.slot(
        type=ConnectionModel, required=True
    )
    devices: cfgdict[DeviceModel] = config.slot(type=DeviceModel, required=True)
    post_prepare: cfglist[typing.Callable[["Simulation", typing.Any], None]] = (
        config.list(type=cfgtypes.function_())
    )

    @staticmethod
    def __plugins__():
        return get_simulation_nodes()

    def get_model_of(
        self, type: typing.Union["CellType", "ConnectionStrategy"]
    ) -> typing.Optional[typing.Union["CellModel", "ConnectionModel"]]:
        cell_models = [cm for cm in self.cell_models.values() if cm.cell_type is type]
        if cell_models:
            return cell_models[0]
        conn_models = [
            cm for cm in self.connection_models.values() if cm.connection_type is type
        ]
        if conn_models:
            return conn_models[0]

    def get_connectivity_sets(
        self,
    ) -> typing.Mapping["ConnectionModel", "ConnectivitySet"]:
        return {
            model: self.scaffold.get_connectivity_set(model.name)
            for model in sorted(self.connection_models.values())
        }


__all__ = ["ProgressEvent", "Simulation"]



"""
Topology module
"""

from ._layout import box_layout
from .partition import AllenStructure, Layer, NrrdVoxels, Partition
from .region import Region, RegionGroup, Stack


def create_topology(regions, ldc, mdc):
    """
    Create a topology from group of regions. Will check for root regions, if there's not
    exactly 1 root region a :class:`~.topology.region.RegionGroup` will be created as new
    root.

    :param regions: Any iterable of regions.
    :type regions: Iterable
    :param ldc: Least dominant corner of the topology. Forms the suggested outer bounds of
      the topology together with the `mdc`.
    :param mdc: Most dominant corner of the topology. Forms the suggested outer bounds of
      the topology together with the `mdc`.
    """
    regions = list(regions)
    if len(roots := get_root_regions(regions)) == 1:
        topology = roots[0]
    else:
        topology = RegionGroup(children=roots, name="topology")
    hint = box_layout(ldc, mdc)
    topology.do_layout(hint)
    return topology


def get_partitions(regions):
    """
    Get all of the partitions belonging to the group of regions and their subregions.

    :param regions: Any iterable of regions.
    :type regions: Iterable
    """

    def collect_deps(region, ignore):
        # get_dependencies can be overridden, so `list` it to avoid mutation of userdata
        deps = list(region.get_dependencies())
        regions = [d for d in deps if not is_partition(d)]
        parts = set(p for p in deps if is_partition(p))
        ignore.update(parts)
        for r in regions:
            # Only check unchecked regions
            if r not in ignore:
                ignore.add(r)
                parts.update(collect_deps(r, ignore))
        return parts

    partitions = set()
    for region in regions:
        partitions.update(p := collect_deps(region, set()))

    return partitions


def is_partition(obj):
    """
    Checks if an object is a partition.
    """
    return (
        hasattr(obj, "get_layout")
        and hasattr(obj, "to_chunks")
        and hasattr(obj, "chunk_to_voxels")
    )


def is_region(obj):
    """
    Checks if an object is a region.
    """
    return hasattr(obj, "get_layout") and hasattr(obj, "do_layout")


def get_root_regions(regions):
    """
    Get all of the root regions, not belonging to any other region in the given group.

    :param regions: Any iterable of regions.
    :type regions: Iterable
    """
    managed = set()

    def collect_deps(region, ignore):
        # get_dependencies can be overridden, so `list` it to avoid mutation of userdata
        deps = list(region.get_dependencies()).copy()
        for dep in deps:
            # Only check unchecked regions, ignore visited & partitions
            if dep not in ignore and is_region(dep):
                ignore.add(dep)
                extra_deps = collect_deps(dep, ignore)
                ignore.update(extra_deps)
        return deps

    # Give `managed` as the mutable ignore arg so that it is filled with all regionals
    # encountered as dependencies by the `collect_deps` recursive function.
    for region in regions:
        collect_deps(region, managed)

    return list(set(list(regions)) - managed)


__all__ = [
    "box_layout",
    "create_topology",
    "get_partitions",
    "get_root_regions",
    "is_partition",
    "is_region",
]



import sys
import warnings


def in_notebook():
    try:
        from IPython import get_ipython

        if "IPKernelApp" not in get_ipython().config:  # pragma: no cover
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


def in_pytest():
    return "pytest" in sys.modules


def report(*message, level=2, ongoing=False, nodes=None, all_nodes=False):
    """
    Send a message to the appropriate output channel.

    :param message: Text message to send.
    :type message: str
    :param level: Verbosity level of the message.
    :type level: int
    :param ongoing: The message is part of an ongoing progress report.
    :type ongoing: bool
    """
    from . import options
    from .services import MPI

    message = " ".join(map(str, message))
    rank = MPI.get_rank()
    if (
        (not rank and nodes is None) or all_nodes or (nodes is not None and rank in nodes)
    ) and options.verbosity >= level:
        print(message, end="\n" if not ongoing else "\r", flush=True)


def warn(message, category=None, stacklevel=2, log_exc=None):
    """
    Send a warning.

    :param message: Warning message
    :type message: str
    :param category: The class of the warning.
    """
    from . import options

    if log_exc:
        import traceback

        from .storage._util import cache

        log = f"{message}\n\n{traceback.format_exception(type(log_exc), log_exc, log_exc.__traceback__)}"
        id = cache.files.store(log)
        path = cache.files.id_to_file_path(id)
        message += f" See '{path}' for full error log."

    # Avoid infinite loop looking up verbosity when verbosity option is broken.
    if "Error retrieving option 'verbosity'" in message or options.verbosity > 0:
        warnings.warn(message, category, stacklevel=stacklevel)


__all__ = [
    "report",
    "warn",
]



import functools
import pathlib
import typing

import appdirs

from ._files import FileDependency

if typing.TYPE_CHECKING:
    from .interfaces import Storage

_bsb_dirs = appdirs.AppDirs("bsb")
_cache_path = pathlib.Path(_bsb_dirs.user_cache_dir)

cache: "Storage"


def __getattr__(name):
    if name == "cache":
        return _get_cache_storage()
    else:
        raise AttributeError(f"{__name__} has no attribute '{name}.")


@functools.cache
def _get_cache_storage():
    from ..storage import Storage

    return Storage("fs", _cache_path)


def _cached_file(source, cache=True):
    return FileDependency(source, file_store=_get_cache_storage().files, cache=cache)



from errr import exception as _e
from errr import make_tree as _t

_t(
    globals(),
    ScaffoldError=_e(
        CodeImportError=_e(),
        CLIError=_e(
            CommandError=_e(),
            ConfigTemplateNotFoundError=_e("template", "path"),
            InputError=_e(),
            DryrunError=_e(),
        ),
        ConfigurationError=_e(
            ConfigurationFormatError=_e(),
            DynamicClassError=_e(
                DynamicObjectNotFoundError=_e(),
                DynamicClassInheritanceError=_e(),
                ClassMapMissingError=_e(),
            ),
            BootError=_e("node"),
            CastError=_e(
                "node",
                "attr",
                DistributionCastError=_e(),
                UnresolvedClassCastError=_e(),
                UnfitClassCastError=_e("node", "attr"),
            ),
            CastConfigurationError=_e(),
            IndicatorError=_e(),
            RequirementError=_e("node", "attr"),
            CfgReferenceError=_e(
                NoReferenceAttributeSignal=_e(),
            ),
            UnknownConfigAttrError=_e("attributes"),
        ),
        CompilationError=_e(
            DistributorError=_e("property", "strategy"),
            RedoError=_e(),
        ),
        DependencyError=_e(),
        GatewayError=_e(
            AllenApiError=_e(),
        ),
        JobPoolError=_e(
            JobCancelledError=_e(),
            JobPoolContextError=_e(),
            JobSchedulingError=_e(),
        ),
        TopologyError=_e(
            UnmanagedPartitionError=_e(),
            LayoutError=_e(),
        ),
        TypeHandlingError=_e(
            NoneReferenceError=_e(),
            InvalidReferenceError=_e("value"),
        ),
        NodeNotFoundError=_e("query"),
        AdapterError=_e(),
        ConnectivityError=_e(
            ExternalSourceError=_e(
                MissingSourceError=_e(),
                IncompleteExternalMapError=_e(),
                SourceQualityError=_e(),
            ),
            UnknownGIDError=_e(),
        ),
        MorphologyError=_e(
            MorphologyRepositoryError=_e(),
            MissingMorphologyError=_e(),
            IncompleteMorphologyError=_e(),
            MorphologyDataError=_e(),
            CircularMorphologyError=_e("morphology", "component").set(list_details=True),
            CompartmentError=_e(),
            EmptySelectionError=_e("selectors"),
            EmptyBranchError=_e(),
        ),
        OptionError=_e(
            ReadOnlyOptionError=_e("option", "tag"),
        ),
        PlacementError=_e(
            ChunkError=_e(),
            PlacementRelationError=_e(),
            ContinuityError=_e(),
            PackingError=_e(),
        ),
        SimulationError=_e(
            ParameterError=_e(
                "parameter",
                ReificationError=_e(),
            )
        ),
        SelectorError=_e(),
        TreeError=_e(),
        VoxelSetError=_e(
            EmptyVoxelSetError=_e(),
        ),
        StorageError=_e(
            DatasetNotFoundError=_e(
                IntersectionDataNotFoundError=_e(),
            ),
            DatasetExistsError=_e(),
            DataNotFoundError=_e(),
            AttributeMissingError=_e(),
            UnknownStorageEngineError=_e(),
            MissingActiveConfigError=_e(),
        ),
        DataNotProvidedError=_e(),
        PluginError=_e("plugin"),
        ParserError=_e(),
        ClassError=_e(),
    ),
)


# Warnings


class ScaffoldWarning(UserWarning):
    pass


class ConfigurationWarning(ScaffoldWarning):
    pass


class PlacementWarning(ScaffoldWarning):
    pass


class PackingWarning(PlacementWarning):
    pass


class MorphologyWarning(ScaffoldWarning):
    pass


class ConnectivityWarning(ScaffoldWarning):
    pass


class PackageRequirementWarning(ScaffoldWarning):
    pass


__all__ = [
    "AdapterError",
    "AllenApiError",
    "AttributeMissingError",
    "BootError",
    "CLIError",
    "CastConfigurationError",
    "CastError",
    "CfgReferenceError",
    "ChunkError",
    "CircularMorphologyError",
    "ClassError",
    "ClassMapMissingError",
    "CodeImportError",
    "CommandError",
    "CompartmentError",
    "CompilationError",
    "ConfigTemplateNotFoundError",
    "ConfigurationError",
    "ConfigurationFormatError",
    "ConfigurationWarning",
    "ConnectivityError",
    "ConnectivityWarning",
    "ContinuityError",
    "DataNotFoundError",
    "DataNotProvidedError",
    "DatasetExistsError",
    "DatasetNotFoundError",
    "DependencyError",
    "DistributionCastError",
    "DistributorError",
    "DryrunError",
    "DynamicClassError",
    "DynamicClassInheritanceError",
    "DynamicObjectNotFoundError",
    "EmptyBranchError",
    "EmptySelectionError",
    "EmptyVoxelSetError",
    "ExternalSourceError",
    "GatewayError",
    "IncompleteExternalMapError",
    "IncompleteMorphologyError",
    "IndicatorError",
    "InputError",
    "IntersectionDataNotFoundError",
    "InvalidReferenceError",
    "JobCancelledError",
    "JobPoolContextError",
    "JobPoolError",
    "JobSchedulingError",
    "LayoutError",
    "MissingActiveConfigError",
    "MissingMorphologyError",
    "MissingSourceError",
    "MorphologyDataError",
    "MorphologyError",
    "MorphologyRepositoryError",
    "MorphologyWarning",
    "NoReferenceAttributeSignal",
    "NodeNotFoundError",
    "NoneReferenceError",
    "OptionError",
    "PackageRequirementWarning",
    "PackingError",
    "PackingWarning",
    "ParameterError",
    "ParserError",
    "PlacementError",
    "PlacementRelationError",
    "PlacementWarning",
    "PluginError",
    "ReadOnlyOptionError",
    "RedoError",
    "ReificationError",
    "RequirementError",
    "ScaffoldError",
    "ScaffoldWarning",
    "SelectorError",
    "SimulationError",
    "SourceQualityError",
    "StorageError",
    "TopologyError",
    "TreeError",
    "TypeHandlingError",
    "UnfitClassCastError",
    "UnknownConfigAttrError",
    "UnknownGIDError",
    "UnknownStorageEngineError",
    "UnmanagedPartitionError",
    "UnresolvedClassCastError",
    "VoxelSetError",
]



import base64
import json
import os
import time
import typing
from uuid import uuid4

from ...exceptions import MissingActiveConfigError
from ..interfaces import FileStore as IFileStore


def _id_to_path(url):
    return base64.urlsafe_b64encode(url.encode("UTF-8")).decode("UTF-8")


def _path_to_id(f):
    return base64.urlsafe_b64decode(f).decode("UTF-8")


class FileStore(IFileStore):
    def file_path(self, *paths):
        return os.path.join(self._engine.root, "files", *paths)

    def id_to_file_path(self, id):
        path = _id_to_path(id)
        return self.file_path(path)

    def meta_path(self, *paths):
        return os.path.join(self._engine.root, "file_meta", *paths)

    def id_to_meta_path(self, id):
        path = _id_to_path(id)
        return self.meta_path(path)

    def all(self):
        return {
            id: self.get_meta(id) for id in map(_path_to_id, os.listdir(self.file_path()))
        }

    def store(self, content, id=None, meta=None, encoding=None, overwrite=False):
        if isinstance(content, str):
            if encoding is None:
                encoding = "utf-8"
            content = content.encode(encoding)
        if id is None:
            id = str(uuid4())
        if meta is None:
            meta = {}
        if not overwrite and self.has(id):
            raise FileExistsError(f"Store already contains a file with id {id}")
        with open(self.id_to_file_path(id), "wb") as f:
            f.write(content)
        with open(self.id_to_meta_path(id), "w") as f:
            json.dump({"meta": meta, "mtime": time.time(), "encoding": encoding}, f)
        return id

    def load(self, id):
        """
        Load the content of an object in the file store.

        :param id: id of the content to be loaded.
        :type id: str
        :returns: The content of the stored object
        :rtype: str
        :raises FileNotFoundError: The given id doesn't exist in the file store.
        """
        with open(self.id_to_file_path(id), "rb") as f:
            content = f.read()
        our_meta = self._get_meta(id)
        meta = our_meta["meta"]
        encoding = our_meta["encoding"]
        return (content.decode(encoding) if encoding else content), meta

    def remove(self, id):
        """
        Remove the content of an object in the file store.

        :param id: id of the content to be removed.
        :type id: str
        :raises FileNotFoundError: The given id doesn't exist in the file store.
        """
        os.unlink(self.id_to_file_path(id))
        os.unlink(self.id_to_meta_path(id))

    def store_active_config(self, config):
        """
        Store configuration in the file store and mark it as the active configuration of
        the stored network.

        :param config: Configuration to be stored
        :type config: :class:`~.config.Configuration`
        :returns: The id the config was stored under
        :rtype: str
        """
        return self.store(json.dumps(config.__tree__()), meta={"active_config": True})

    def load_active_config(self):
        """
        Load the active configuration stored in the file store.

        :returns: The active configuration
        :rtype: :class:`~.config.Configuration`
        :raises Exception: When there's no active configuration in the file store.
        """
        from ...config import Configuration

        stored = self.find_meta("active_config", True)
        if stored is None:
            raise MissingActiveConfigError("No active config")
        else:
            content, meta = stored.load()
            tree = json.loads(content)
            cfg = Configuration(**tree, _store=self)
            cfg._meta = meta
            return cfg

    def has(self, id):
        """
        Must return whether the file store has a file with the given id.
        """
        return os.path.exists(self.id_to_file_path(id))

    def get_mtime(self, id):
        """
        Must return the last modified timestamp of file with the given id.
        """
        return self._get_meta(id)["mtime"]

    def get_encoding(self, id):
        """
        Must return the encoding of the file with the given id, or None if it is
        unspecified binary data.
        """
        return self._get_meta(id)["encoding"]

    def _get_meta(self, id) -> typing.Mapping[str, typing.Any]:
        """
        Must return the metadata of the given id.
        """
        with open(self.id_to_meta_path(id), "r") as f:
            return json.load(f)

    def get_meta(self, id) -> typing.Mapping[str, typing.Any]:
        """
        Must return the metadata of the given id.
        """
        return self._get_meta(id)["meta"]



import pathlib

import toml

from ... import config
from ...reporting import report
from . import BaseCommand


class ProjectNewCommand(BaseCommand, name="new"):
    def get_options(self):
        return {}

    def add_parser_arguments(self, parser):
        parser.add_argument("project_name", nargs="?", help="Project name", default="")
        parser.add_argument(
            "path", nargs="?", default=".", help="Location of the project"
        )
        parser.add_argument(
            "--quickstart", action="store_true", help="Start an example project"
        )
        parser.add_argument(
            "--json", action="store_true", help="Use JSON as configuration language"
        )
        parser.add_argument(
            "--exists",
            action="store_true",
            help="Indicates whether the folder structure already exists.",
        )

    def handler(self, context):
        name = (
            context.arguments.project_name
            or input("Project name [my_model]: ")
            or "my_model"
        )
        root = pathlib.Path(context.arguments.path) / name
        try:
            root.mkdir(exist_ok=context.arguments.exists)
        except FileExistsError:
            return report(
                f"Could not create '{root.absolute()}', directory exists.", level=0
            )
        ext = "json" if context.arguments.json else "yaml"
        if context.arguments.quickstart:
            template = f"starting_example.{ext}"
            output = f"network_configuration.{ext}"
        else:
            template = input(f"Config template [skeleton.{ext}]: ") or f"skeleton.{ext}"
            output = (
                input(f"Config filename [network_configuration.{ext}]: ")
                or f"network_configuration.{ext}"
            )
        config.copy_template(template, output=root / output)
        with open(root / "pyproject.toml", "w") as f:
            toml.dump(
                {
                    "tools": {
                        "bsb": {
                            "config": output,
                        }
                    }
                },
                f,
            )
        place_path = root / "placement.py"
        conn_path = root / "connectome.py"
        if not place_path.exists():
            with open(place_path, "w") as f:
                f.write("from bsb import PlacementStrategy\n")
        if not conn_path.exists():
            with open(conn_path, "w") as f:
                f.write("from bsb import ConnectionStrategy\n")

        report(f"Created '{name}' project structure.", level=1)



import os
import shutil
from datetime import datetime

import shortuuid

from ... import config
from ...services import MPILock
from ..decorators import on_main, on_main_until
from ..interfaces import Engine, NoopLock
from ..interfaces import StorageNode as IStorageNode
from .file_store import FileStore


class FileSystemEngine(Engine):
    def __init__(self, root, comm):
        super().__init__(root, comm)
        self._lock = MPILock.sync()
        self._readonly = False

    @property
    def root_slug(self):
        return os.path.relpath(self._root)

    @classmethod
    def recognizes(cls, root):
        try:
            return os.path.exists(root) and os.path.isdir(root)
        except Exception:
            return False

    def _read(self):
        if self._readonly:
            return NoopLock()
        else:
            return self._lock.read()

    def _write(self):
        if self._readonly:
            raise IOError("Can't perform write operations in readonly mode.")
        else:
            return self._lock.write()

    def _master_write(self):
        if self._readonly:
            raise IOError("Can't perform write operations in readonly mode.")
        else:
            return self._lock.single_write()

    def exists(self):
        return os.path.exists(self._root)

    @on_main_until(lambda self: self.exists())
    def create(self):
        os.makedirs(os.path.join(self._root, "files"), exist_ok=True)
        os.makedirs(os.path.join(self._root, "file_meta"), exist_ok=True)

    @on_main_until(lambda self: self.exists())
    def move(self, new_root):
        shutil.move(self._root, new_root)
        self._root = new_root

    @on_main_until(lambda self, r: self.__class__(self.root, self.comm).exists())
    def copy(self, new_root):
        shutil.copytree(self._root, new_root)

    @on_main_until(lambda self: not self.exists())
    def remove(self):
        shutil.rmtree(self._root)

    def require_placement_set(self, ct):
        raise NotImplementedError("No PS")

    def clear_placement(self):
        pass

    def clear_connectivity(self):
        pass

    def get_chunk_stats(self):
        return {}


def _get_default_root():
    return os.path.abspath(
        os.path.join(
            ".",
            "scaffold_network_"
            + datetime.now().strftime("%Y_%m_%d")
            + "_"
            + shortuuid.uuid(),
        )
    )


@config.node
class StorageNode(IStorageNode):
    root = config.attr(type=str, default=_get_default_root, call_default=True)
    """
    Path to the filesystem storage file.
    """



import itertools
import os
import sys
import typing

import numpy as np

from ._util import obj_str_insert
from .config._config import Configuration
from .connectivity import ConnectionStrategy
from .exceptions import (
    InputError,
    MissingActiveConfigError,
    NodeNotFoundError,
    RedoError,
)
from .placement import PlacementStrategy
from .profiling import meter
from .reporting import report
from .services import MPI, JobPool
from .services._pool_listeners import NonTTYTerminalListener, TTYTerminalListener
from .services.pool import Job, Workflow
from .simulation import get_simulation_adapter
from .storage import Storage, open_storage
from .storage._chunks import Chunk

if typing.TYPE_CHECKING:
    from .cell_types import CellType
    from .config._config import NetworkNode as Network
    from .postprocessing import AfterPlacementHook
    from .simulation.simulation import Simulation
    from .storage.interfaces import (
        ConnectivitySet,
        FileStore,
        MorphologyRepository,
        PlacementSet,
    )
    from .topology import Partition, Region


@meter()
def from_storage(root):
    """
    Load :class:`.core.Scaffold` from a storage object.

    :param root: Root (usually path) pointing to the storage object.
    :returns: A network scaffold
    :rtype: :class:`Scaffold`
    """
    return open_storage(root).load()


_cfg_props = (
    "network",
    "regions",
    "partitions",
    "cell_types",
    "placement",
    "after_placement",
    "connectivity",
    "after_connectivity",
    "simulations",
)


def _config_property(name):
    def fget(self):
        return getattr(self.configuration, name)

    def fset(self, value):
        setattr(self.configuration, name, value)

    prop = property(fget)
    return prop.setter(fset)


def _get_linked_config(storage=None):
    import bsb.config

    try:
        cfg = storage.load_active_config()
    except Exception:
        import bsb.options

        path = bsb.options.config
    else:
        path = cfg._meta.get("path", None)
    if path and os.path.exists(path):
        with open(path, "r") as f:
            cfg = bsb.config.parse_configuration_file(f)
            return cfg
    else:
        return None


def _bad_flag(flag: bool):
    return flag is not None and bool(flag) is not flag


class Scaffold:
    """

    This is the main object of the bsb package, it represents a network and puts together
    all the pieces that make up the model description such as the
    :class:`~.config.Configuration` with the technical side like the
    :class:`~.storage.Storage`.
    """

    network: "Network"
    regions: typing.Dict[str, "Region"]
    partitions: typing.Dict[str, "Partition"]
    cell_types: typing.Dict[str, "CellType"]
    placement: typing.Dict[str, "PlacementStrategy"]
    after_placement: typing.Dict[str, "AfterPlacementHook"]
    connectivity: typing.Dict[str, "ConnectionStrategy"]
    after_connectivity: typing.Dict[str, "AfterPlacementHook"]
    simulations: typing.Dict[str, "Simulation"]

    def __init__(self, config=None, storage=None, clear=False, comm=None):
        """
        Bootstraps a network object.

        :param config: The configuration to use for this network. If it is omitted the
          :ref:`default configuration <default-config>` is used.
        :type config: :class:`~.config.Configuration`
        :param storage: The storage to use to read and write data for this network. If it
          is omitted the configuration's ``Storage`` node is used to construct one.
        :type storage: :class:`~.storage.Storage`
        :param clear: Start with a new network, clearing any previously stored information
        :type clear: bool
        :returns: A network object
        :rtype: :class:`~.core.Scaffold`
        """
        self._pool_listeners: list[tuple[typing.Callable[[list["Job"]], None], float]] = (
            []
        )
        self._configuration = None
        self._storage = None
        self._comm = comm or MPI
        self._bootstrap(config, storage, clear=clear)

    def __contains__(self, component):
        return getattr(component, "scaffold", None) is self

    @obj_str_insert
    def __repr__(self):
        file = os.path.abspath(self.storage.root)
        cells_placed = len(self.cell_types)
        n_types = len(self.connectivity)
        return f"'{file}' with {cells_placed} cell types, and {n_types} connection_types"

    def is_main_process(self) -> bool:
        return not MPI.get_rank()

    def is_worker_process(self) -> bool:
        return bool(MPI.get_rank())

    def _bootstrap(self, config, storage, clear=False):
        if config is None:
            # No config given, check for linked configs, or stored configs, otherwise
            # make default config.
            linked = _get_linked_config(storage)
            if linked:
                report(f"Pulling configuration from linked {linked}.", level=2)
                config = linked
            elif storage is not None:
                try:
                    config = storage.load_active_config()
                except MissingActiveConfigError:
                    config = Configuration.default()
            else:
                config = Configuration.default()
        if not storage:
            # No storage given, create one.
            report("Creating storage from config.", level=4)
            storage = Storage(config.storage.engine, config.storage.root)
        if clear:
            # Storage given, but asked to clear it before use.
            storage.remove()
            storage.create()
        # Synchronize the scaffold, config and storage objects for use together
        self._configuration = config
        # Make sure the storage config node reflects the storage we are using
        config._update_storage_node(storage)
        # Give the scaffold access to the unitialized storage object (for use during
        # config bootstrapping).
        self._storage = storage
        # First, the scaffold is passed to each config node, and their boot methods called
        self._configuration._bootstrap(self)
        # Then, `storage` is initted for the scaffold, and `config` is stored (happens
        # inside the `storage` property).
        self.storage = storage

    storage_cfg = _config_property("storage")
    for attr in _cfg_props:
        vars()[attr] = _config_property(attr)

    @property
    def configuration(self) -> Configuration:
        return self._configuration

    @configuration.setter
    def configuration(self, cfg: Configuration):
        self._configuration = cfg
        cfg._bootstrap(self)
        self.storage.store_active_config(cfg)

    @property
    def storage(self) -> Storage:
        return self._storage

    @storage.setter
    def storage(self, storage: Storage):
        self._storage = storage
        storage.init(self)

    @property
    def morphologies(self) -> "MorphologyRepository":
        return self.storage.morphologies

    @property
    def files(self) -> "FileStore":
        return self.storage.files

    def clear(self):
        """
        Clears the storage. This deletes any existing network data!
        """
        self.storage.renew(self)

    def clear_placement(self):
        """
        Clears the placement storage.
        """
        self.storage.clear_placement(self)

    def clear_connectivity(self):
        """
        Clears the connectivity storage.
        """
        self.storage.clear_connectivity()

    def resize(self, x=None, y=None, z=None):
        """
        Updates the topology boundary indicators. Use before placement, updates
        only the abstract topology tree, does not rescale, prune or otherwise
        alter already existing placement data.
        """
        from .topology._layout import box_layout

        if x is not None:
            self.network.x = x
        if y is not None:
            self.network.y = y
        if z is not None:
            self.network.z = z
        self.topology.do_layout(
            box_layout(
                self.network.origin,
                np.array(self.network.origin)
                + [self.network.x, self.network.y, self.network.z],
            )
        )

    @meter()
    def run_placement(self, strategies=None, fail_fast=True, pipelines=True):
        """
        Run placement strategies.
        """
        if pipelines:
            self.run_pipelines()
        if strategies is None:
            strategies = [*self.placement.values()]
        strategies = PlacementStrategy.sort_deps(strategies)
        with self.create_job_pool(fail_fast=fail_fast) as pool:
            if pool.is_main():

                def scheduler(strategy):
                    strategy.queue(pool, self.network.chunk_size)

                pool.schedule(strategies, scheduler)
            pool.execute()

    @meter()
    def run_connectivity(self, strategies=None, fail_fast=True, pipelines=True):
        """
        Run connection strategies.
        """
        if pipelines:
            self.run_pipelines()
        if strategies is None:
            strategies = set(self.connectivity.values())
        strategies = ConnectionStrategy.sort_deps(strategies)
        with self.create_job_pool(fail_fast=fail_fast) as pool:
            if pool.is_main():
                pool.schedule(strategies)
            pool.execute()

    @meter()
    def run_placement_strategy(self, strategy):
        """
        Run a single placement strategy.
        """
        self.run_placement([strategy])

    @meter()
    def run_after_placement(self, hooks=None, fail_fast=None, pipelines=True):
        """
        Run after placement hooks.
        """
        if hooks is None:
            hooks = self.after_placement
        with self.create_job_pool(fail_fast) as pool:
            if pool.is_main():
                pool.schedule(hooks)
            pool.execute()

    @meter()
    def run_after_connectivity(self, hooks=None, fail_fast=None, pipelines=True):
        """
        Run after placement hooks.
        """
        if hooks is None:
            hooks = self.after_placement
        with self.create_job_pool(fail_fast) as pool:
            if pool.is_main():
                pool.schedule(hooks)
            pool.execute()

    @meter()
    def compile(
        self,
        skip_placement=False,
        skip_connectivity=False,
        skip_after_placement=False,
        skip_after_connectivity=False,
        only=None,
        skip=None,
        clear=False,
        append=False,
        redo=False,
        force=False,
        fail_fast=True,
    ):
        """
        Run reconstruction steps in the scaffold sequence to obtain a full network.
        """
        existed = self.storage.preexisted
        if skip_placement:
            p_strats = []
        else:
            p_strats = self.get_placement(skip=skip, only=only)
        if skip_connectivity:
            c_strats = []
        else:
            c_strats = self.get_connectivity(skip=skip, only=only)
        todo_list_str = ", ".join(s.name for s in itertools.chain(p_strats, c_strats))
        report(f"Compiling the following strategies: {todo_list_str}", level=2)
        if _bad_flag(clear) or _bad_flag(redo) or _bad_flag(append):
            raise InputError(
                "`clear`, `redo` and `append` are strictly boolean flags. "
                "Pass the strategies to run to the skip/only options instead."
            )
        if sum((bool(clear), bool(redo), bool(append))) > 1:
            raise InputError("`clear`, `redo` and `append` are mutually exclusive.")
        if existed:
            if not (clear or append or redo):
                raise FileExistsError(
                    f"The `{self.storage.format}` storage"
                    + f" at `{self.storage.root}` already exists. Either move/delete it,"
                    + " or pass one of the `clear`, `append` or `redo` arguments"
                    + " to pick what to do with the existing data."
                )
            if clear:
                report("Clearing data", level=2)
                # Clear the placement and connectivity data, but leave any cached files
                # and morphologies intact.
                self.clear_placement()
                self.clear_connectivity()
            elif redo:
                # In order to properly redo things, we clear some placement and connection
                # data, but since multiple placement/connection strategies can contribute
                # to the same sets we might be wiping their data too, and they will need
                # to be cleared and reran as well.
                p_strats, c_strats = self._redo_chain(p_strats, c_strats, skip, force)
            # else:
            #   append mode is luckily simpler, just don't clear anything :)

        phases = ["pipelines"]
        if not skip_placement:
            phases.append("placement")
        if not skip_after_placement:
            phases.append("after_placement")
        if not skip_connectivity:
            phases.append("connectivity")
        if not skip_after_connectivity:
            phases.append("after_connectivity")
        self._workflow = Workflow(phases)
        try:
            self.run_pipelines(fail_fast=fail_fast)
            self._workflow.next_phase()
            if not skip_placement:
                placement_todo = ", ".join(s.name for s in p_strats)
                report(f"Starting placement strategies: {placement_todo}", level=2)
                self.run_placement(p_strats, fail_fast=fail_fast, pipelines=False)
                self._workflow.next_phase()
            if not skip_after_placement:
                self.run_after_placement(pipelines=False, fail_fast=fail_fast)
                self._workflow.next_phase()
            if not skip_connectivity:
                connectivity_todo = ", ".join(s.name for s in c_strats)
                report(f"Starting connectivity strategies: {connectivity_todo}", level=2)
                self.run_connectivity(c_strats, fail_fast=fail_fast, pipelines=False)
                self._workflow.next_phase()
            if not skip_after_connectivity:
                self.run_after_connectivity(pipelines=False)
                self._workflow.next_phase()
        finally:
            # After compilation we should flag the storage as having existed before so that
            # the `clear`, `redo` and `append` flags take effect on a second `compile` pass.
            self.storage._preexisted = True
            del self._workflow

    @meter()
    def run_pipelines(self, fail_fast=True, pipelines=None):
        if pipelines is None:
            pipelines = self.get_dependency_pipelines()
        with self.create_job_pool(fail_fast=fail_fast) as pool:
            if pool.is_main():
                pool.schedule(pipelines)
            pool.execute()

    @meter()
    def run_simulation(self, simulation_name: str):
        """
        Run a simulation starting from the default single-instance adapter.

        :param simulation_name: Name of the simulation in the configuration.
        :type simulation_name: str
        """
        simulation = self.get_simulation(simulation_name)
        adapter = get_simulation_adapter(simulation.simulator)
        return adapter.simulate(simulation)[0]

    def get_simulation(self, sim_name: str) -> "Simulation":
        """
        Retrieve the default single-instance adapter for a simulation.
        """
        if sim_name not in self.simulations:
            simstr = ", ".join(f"'{s}'" for s in self.simulations.keys())
            raise NodeNotFoundError(
                f"Unknown simulation '{sim_name}', choose from: {simstr}"
            )
        return self.configuration.simulations[sim_name]

    def place_cells(
        self,
        cell_type,
        positions,
        morphologies=None,
        rotations=None,
        additional=None,
        chunk=None,
    ):
        """
        Place cells inside of the scaffold

        .. code-block:: python

            # Add one granule cell at position 0, 0, 0
            cell_type = scaffold.get_cell_type("granule_cell")
            scaffold.place_cells(cell_type, cell_type.layer_instance, [[0., 0., 0.]])

        :param cell_type: The type of the cells to place.
        :type cell_type: ~bsb.cell_types.CellType
        :param positions: A collection of xyz positions to place the cells on.
        :type positions: Any `np.concatenate` type of shape (N, 3).
        """
        if chunk is None:
            chunk = Chunk([0, 0, 0], self.network.chunk_size)
        if hasattr(chunk, "dimensions") and np.any(np.isnan(chunk.dimensions)):
            chunk.dimensions = self.network.chunk_size
        self.get_placement_set(cell_type).append_data(
            chunk,
            positions=positions,
            morphologies=morphologies,
            rotations=rotations,
            additional=additional,
        )

    def create_entities(self, cell_type, count):
        """
        Create entities in the simulation space.

        Entities are different from cells because they have no positional data and
        don't influence the placement step. They do have a representation in the
        connection and simulation step.

        :param cell_type: The cell type of the entities
        :type cell_type: ~bsb.cell_types.CellType
        :param count: Number of entities to place
        :type count: int
        :todo: Allow `additional` data for entities
        """
        if count == 0:
            return
        ps = self.get_placement_set(cell_type)
        # Append entity data to the default chunk 000
        chunk = Chunk([0, 0, 0], self.network.chunk_size)
        ps.append_entities(chunk, count)

    def get_placement(
        self, cell_types=None, skip=None, only=None
    ) -> typing.List["PlacementStrategy"]:
        if cell_types is not None:
            cell_types = [
                self.cell_types[ct] if isinstance(ct, str) else ct for ct in cell_types
            ]
        return [
            val
            for key, val in self.placement.items()
            if (cell_types is None or any(ct in cell_types for ct in val.cell_types))
            and (only is None or key in only)
            and (skip is None or key not in skip)
        ]

    def get_placement_of(self, *cell_types):
        """
        Find all of the placement strategies that given certain cell types.

        :param cell_types: Cell types (or their names) of interest.
        :type cell_types: Union[~bsb.cell_types.CellType, str]
        """
        return self.get_placement(cell_types=cell_types)

    def get_placement_set(
        self, type, chunks=None, labels=None, morphology_labels=None
    ) -> "PlacementSet":
        """
        Return a cell type's placement set from the output formatter.

        :param tag: Unique identifier of the placement set in the storage
        :type tag: str
        :returns: A placement set
        :param labels: Labels to filter the placement set by.
        :type labels: list[str]
        :param morphology_labels: Subcellular labels to apply to the morphologies.
        :type morphology_labels: list[str]
        :rtype: :class:`~.storage.interfaces.PlacementSet`
        """
        if isinstance(type, str):
            type = self.cell_types[type]
        return self.storage.get_placement_set(
            type, chunks=chunks, labels=labels, morphology_labels=morphology_labels
        )

    def get_placement_sets(self) -> typing.List["PlacementSet"]:
        """
        Return all of the placement sets present in the network.

        :rtype: List[~bsb.storage.interfaces.PlacementSet]
        """
        return [cell_type.get_placement_set() for cell_type in self.cell_types.values()]

    def get_connectivity(
        self, anywhere=None, presynaptic=None, postsynaptic=None, skip=None, only=None
    ) -> typing.List["ConnectivitySet"]:
        conntype_filtered = self._connectivity_query(
            any_query=set(self._sanitize_ct(anywhere)),
            pre_query=set(self._sanitize_ct(presynaptic)),
            post_query=set(self._sanitize_ct(postsynaptic)),
        )
        return [
            ct
            for ct in conntype_filtered
            if (only is None or ct.name in only) and (skip is None or ct.name not in skip)
        ]

    def get_connectivity_sets(self) -> typing.List["ConnectivitySet"]:
        """
        Return all connectivity sets from the output formatter.

        :param tag: Unique identifier of the connectivity set in the output formatter
        :type tag: str
        :returns: All connectivity sets
        """
        return [self._load_cs_types(cs) for cs in self.storage.get_connectivity_sets()]

    def require_connectivity_set(self, pre, post, tag=None) -> "ConnectivitySet":
        return self._load_cs_types(
            self.storage.require_connectivity_set(pre, post, tag), pre, post
        )

    def get_connectivity_set(self, tag=None, pre=None, post=None) -> "ConnectivitySet":
        """
        Return a connectivity set from the output formatter.

        :param tag: Unique identifier of the connectivity set in the output formatter
        :type tag: str
        :returns: A connectivity set
        :rtype: :class:`~.storage.interfaces.ConnectivitySet`
        """
        if tag is None:
            try:
                tag = f"{pre.name}_to_{post.name}"
            except Exception:
                raise ValueError("Supply either `tag` or a valid pre and post cell type.")
        return self._load_cs_types(self.storage.get_connectivity_set(tag), pre, post)

    def get_cell_types(self) -> typing.List["CellType"]:
        """
        Return a list of all cell types in the network.
        """
        return [*self.configuration.cell_types.values()]

    def merge(self, other, label=None):
        raise NotImplementedError("Revisit: merge CT, PS & CS, done?")

    def _sanitize_ct(self, seq_str_or_none):
        if seq_str_or_none is None:
            return []
        try:
            if isinstance(seq_str_or_none, str):
                return [self.cell_types[seq_str_or_none]]
            return [
                self.cell_types[s] if isinstance(s, str) else s for s in seq_str_or_none
            ]
        except KeyError as e:
            raise NodeNotFoundError(f"Cell type `{e.args[0]}` not found.")

    def _connectivity_query(self, any_query=set(), pre_query=set(), post_query=set()):
        # Filter network connection types for any type that satisfies both
        # the presynaptic and postsynaptic query. Empty queries satisfy all
        # types. The presynaptic query is satisfied if the conn type contains
        # any of the queried cell types presynaptically, and same for post.
        # The any query is satisfied if a cell type is found either pre or post.

        def partial_query(types, query):
            return not query or any(cell_type in query for cell_type in types)

        def query(conn_type):
            pre_match = partial_query(conn_type.presynaptic.cell_types, pre_query)
            post_match = partial_query(conn_type.postsynaptic.cell_types, post_query)
            any_match = partial_query(
                conn_type.presynaptic.cell_types, any_query
            ) or partial_query(conn_type.postsynaptic.cell_types, any_query)
            return any_match or (pre_match and post_match)

        types = self.connectivity.values()
        return [*filter(query, types)]

    def _redo_chain(self, p_strats, c_strats, skip, force):
        p_contrib = set(p_strats)
        while True:
            # Get all the placement strategies that effect the current set of CT.
            full_wipe = set(itertools.chain(*(ps.cell_types for ps in p_contrib)))
            contrib = set(self.get_placement(full_wipe))
            # Keep repeating until no new contributors are fished up.
            if contrib.issubset(p_contrib):
                break
            # Grow the placement chain
            p_contrib.update(contrib)
        report(
            "Redo-affected placement: " + " ".join(ps.name for ps in p_contrib), level=2
        )

        c_contrib = set(c_strats)
        conn_wipe = full_wipe.copy()
        if full_wipe:
            while True:
                contrib = set(self.get_connectivity(anywhere=conn_wipe))
                conn_wipe.update(
                    itertools.chain(*(ct.get_cell_types() for ct in contrib))
                )
                if contrib.issubset(c_contrib):
                    break
                c_contrib.update(contrib)
        report(
            "Redo-affected connectivity: " + " ".join(cs.name for cs in c_contrib),
            level=2,
        )
        # Don't do greedy things without `force`
        if not force:
            # Error if we need to redo things the user asked to skip
            if skip is not None:
                unskipped = [p.name for p in p_contrib if p.name in skip]
                if unskipped:
                    chainstr = ", ".join(f"'{s.name}'" for s in (p_strats + c_strats))
                    skipstr = ", ".join(f"'{s.name}'" for s in unskipped)
                    raise RedoError(
                        f"Can't skip {skipstr}. Redoing {chainstr} requires to redo them."
                        + f" Omit {skipstr} from `skip` or use `force` (not recommended)."
                    )
            # Error if we need to redo things the user didn't ask for
            for label, chain, og in zip(
                ("placement", "connection"), (p_contrib, c_contrib), (p_strats, c_strats)
            ):
                if len(chain) > len(og):
                    new = chain.difference(og)
                    raise RedoError(
                        f"Need to redo additional {label} strategies: "
                        + ", ".join(n.name for n in new)
                        + ". Include them or use `force` (not recommended)."
                    )

        for ct in full_wipe:
            report(f"Clearing all data of {ct.name}", level=2)
            ct.clear()

        for ct in conn_wipe:
            report(f"Clearing connectivity data of {ct.name}", level=2)
            ct.clear_connections()

        return p_contrib, c_contrib

    def get_dependency_pipelines(self):
        return [*self.configuration.morphologies]

    def get_config_diagram(self):
        from .config import make_configuration_diagram

        return make_configuration_diagram(self.configuration)

    def get_storage_diagram(self):
        dot = f'digraph "{self.configuration.name or "network"}" {{'
        for ps in self.get_placement_sets():
            dot += f'\n  {ps.tag}[label="{ps.tag} ({len(ps)} {ps.cell_type.name})"]'
        for conn in self.get_connectivity_sets():
            dot += f"\n  {conn.pre_type.name} -> {conn.post_type.name}"
            dot += f'[label="{conn.tag} ({len(conn)})"];'

        dot += "\n}\n"
        return dot

    def _load_cs_types(
        self, cs: "ConnectivitySet", pre=None, post=None
    ) -> "ConnectivitySet":
        if pre and pre.name != cs.pre_type_name:
            raise ValueError(
                "Given and stored type mismatch:" + f" {pre.name} vs {cs.pre_type_name}"
            )
        if post and post.name != cs.post_type_name:
            raise ValueError(
                "Given and stored type mismatch:" + f" {post.name} vs {cs.post_type_name}"
            )
        try:
            cs.pre_type = self.cell_types[cs.pre_type_name]
            cs.post_type = self.cell_types[cs.post_type_name]
        except KeyError as e:
            raise NodeNotFoundError(
                f"Couldn't load '{cs.tag}' connections, missing cell type '{e.args[0]}'."
            ) from None
        return cs

    def create_job_pool(self, fail_fast=None, quiet=False):
        pool = JobPool(
            self, fail_fast=fail_fast, workflow=getattr(self, "_workflow", None)
        )
        try:
            # Check whether stdout is a TTY, and that it is larger than 0x0
            # (e.g. MPI sets it to 0x0 unless an xterm is emulated.
            tty = os.isatty(sys.stdout.fileno()) and sum(os.get_terminal_size())
        except Exception:
            tty = False
        if tty:
            fps = 25
            default_listener = TTYTerminalListener(fps)
            default_max_wait = 1 / fps
        else:
            default_listener = NonTTYTerminalListener()
            default_max_wait = None
        if self._pool_listeners:
            for listener, max_wait in self._pool_listeners:
                pool.add_listener(listener, max_wait=max_wait)
        elif not quiet:
            pool.add_listener(default_listener, max_wait=default_max_wait)
        return pool

    def register_listener(self, listener, max_wait=None):
        self._pool_listeners.append((listener, max_wait))

    def remove_listener(self, listener):
        for i, (l, _) in enumerate(self._pool_listeners):
            if l is listener:
                self._pool_listeners.pop(i)
                break


class ReportListener:
    def __init__(self, scaffold, file):
        self.file = file
        self.scaffold = scaffold

    def __call__(self, progress):
        report(
            str(progress.progression)
            + "+"
            + str(progress.duration)
            + "+"
            + str(progress.time),
            token="simulation_progress",
        )


__all__ = ["ReportListener", "Scaffold", "from_storage"]



import typing

import numpy as np

from .. import config
from .._util import merge_dicts
from ..cell_types import CellType
from ..connectivity import ConnectionStrategy
from ..placement import PlacementStrategy
from ..postprocessing import AfterPlacementHook
from ..simulation.simulation import Simulation
from ..storage._files import (
    CodeDependencyNode,
    MorphologyDependencyNode,
    MorphologyPipelineNode,
)
from ..storage.interfaces import StorageNode
from ..topology import Partition, Region, RegionGroup, create_topology, get_partitions
from . import types
from ._attrs import _boot_nodes, cfgdict, cfglist

if typing.TYPE_CHECKING:
    import packaging.requirements

    from ..core import Scaffold


@config.node
class NetworkNode:
    scaffold: "Scaffold"

    x: float = config.attr(type=float, required=True)
    y: float = config.attr(type=float, required=True)
    z: float = config.attr(type=float, required=True)
    origin: list[float] = config.attr(
        type=types.list(type=float, size=3),
        default=lambda: [0, 0, 0],
        call_default=True,
    )
    chunk_size: list[float] = config.attr(
        type=types.or_(
            types.list(float),
            types.scalar_expand(float, expand=lambda s: np.ones(3) * s),
        ),
        default=lambda: [100.0, 100.0, 100.0],
        call_default=True,
    )

    def boot(self):
        self.chunk_size = np.array(self.chunk_size)


@config.root
class Configuration:
    """
    The main Configuration object containing the full definition of a scaffold model.
    """

    scaffold: "Scaffold"

    name: str = config.attr()
    """
    Descriptive name of the model
    """
    components: cfglist[CodeDependencyNode] = config.list(
        type=CodeDependencyNode,
    )
    """
    List of modules relative to the project root containing extension components.
    """
    packages: cfglist["packaging.requirements.Requirement"] = config.list(
        type=types.PackageRequirement(),
    )
    """
    List of package requirement specifiers the model depends on.
    
    """
    morphologies: cfglist[MorphologyDependencyNode] = config.list(
        type=types.or_(MorphologyDependencyNode, MorphologyPipelineNode),
    )
    """
    Morphology files and processing pipelines.
    """
    storage: StorageNode = config.attr(
        type=StorageNode,
        required=True,
    )
    """
    Network storage configuration
    """
    network: NetworkNode = config.attr(
        type=NetworkNode,
        required=True,
    )
    """
    Network description
    """
    regions: cfgdict[str, Region] = config.dict(
        type=Region,
    )
    """
    Network regions
    """
    partitions: cfgdict[str, Partition] = config.dict(
        type=Partition,
        required=True,
    )
    """
    Network partitions
    """
    cell_types: cfgdict[str, CellType] = config.dict(
        type=CellType,
        required=True,
    )
    """
    Network cell types
    """
    placement: cfgdict[str, PlacementStrategy] = config.dict(
        type=PlacementStrategy,
        required=True,
    )
    """
    Network placement strategies
    """
    after_placement: cfgdict[str, AfterPlacementHook] = config.dict(
        type=AfterPlacementHook,
    )
    connectivity: cfgdict[str, ConnectionStrategy] = config.dict(
        type=ConnectionStrategy,
        required=True,
    )
    """
    Network connectivity strategies
    """
    after_connectivity: cfgdict[str, AfterPlacementHook] = config.dict(
        type=AfterPlacementHook,
    )
    simulations: cfgdict[str, Simulation] = config.dict(
        type=Simulation,
    )
    """
    Network simulation configuration
    """
    __module__ = "bsb.config"

    @classmethod
    def default(cls, **kwargs):
        default_args = dict(
            storage={"engine": "fs"},
            network={"x": 200, "y": 200, "z": 200},
            partitions={},
            cell_types={},
            placement={},
            connectivity={},
        )
        merge_dicts(default_args, kwargs)
        conf = cls(default_args)
        conf._parser = "json"
        conf._file = None
        return conf

    def _bootstrap(self, scaffold):
        # Activate the scaffold property of each config node
        _boot_nodes(self, scaffold)
        self._config_isbooted = True
        # Initialise the topology from the defined regions
        regions = list(self.regions.values())
        # Arrange the topology based on network boundaries
        start = self.network.origin.copy()
        net = self.network
        end = [start[0] + net.x, start[1] + net.y, start[2] + net.z]
        # If there are any partitions not part of the topology, add them to a group
        if unmanaged := set(self.partitions.values()) - get_partitions(regions):
            p = "', '".join(p.name for p in unmanaged)
            r = scaffold.regions.add(
                "__unmanaged__", RegionGroup(children=list(unmanaged))
            )
            regions.append(r)
        scaffold.topology = create_topology(regions, start, end)

    def _update_storage_node(self, storage):
        if self.storage.engine != storage.format:
            self.storage.engine = storage.format
        if self.storage.root != storage.root:
            self.storage.root = storage.root

    def __str__(self):
        return str(self.__tree__())

    def __repr__(self):
        return f"{type(self).__qualname__}({self})"



import typing

import errr
import numpy as np
import scipy.stats.distributions as _distributions

from .. import config
from ..exceptions import DistributionCastError
from . import types

if typing.TYPE_CHECKING:
    from ..core import Scaffold

# Scan the scipy distributions module for all distribution names. Ignore `_gen` which are
# the factory functions for the distribution classes. `rvs` is a duck type check.
_available_distributions = [
    d
    for d, v in _distributions.__dict__.items()
    if hasattr(v, "rvs") and not d.endswith("_gen")
]
_available_distributions.append("constant")


@config.node
class Distribution:
    scaffold: "Scaffold"
    distribution: str = config.attr(
        type=types.in_(_available_distributions), required=True
    )
    parameters: dict[str, typing.Any] = config.catch_all(type=types.any_())

    def __init__(self, **kwargs):
        if self.distribution == "constant":
            self._distr = _ConstantDistribution(self.parameters["constant"])
            return

        try:
            self._distr = getattr(_distributions, self.distribution)(**self.parameters)
        except Exception as e:
            errr.wrap(
                DistributionCastError, e, prepend=f"Can't cast to '{self.distribution}': "
            )

    def draw(self, n):
        return self._distr.rvs(size=n)

    def __getattr__(self, attr):
        if "_distr" not in self.__dict__:
            raise AttributeError("No underlying _distr found for distribution node.")
        return getattr(self._distr, attr)


class _ConstantDistribution:
    def __init__(self, const):
        self.const = const

    def rvs(self, size):
        return np.full(size, self.const, dtype=type(self.const))



# isort: off
# Load module before others to prevent partially initialized modules
from .strategy import ConnectionStrategy

# isort: on
from .detailed import *
from .general import *
from .import_ import CsvImportConnectivity
